"""
Provides a DVRWavefunction class that inherits from the base Psience wavefunction
"""

import numpy as np

from McUtils.Zachary import Mesh
import McUtils.Plots as plt

from Psience.Wavefun import Wavefunction, Wavefunctions
from .Gaussians import DGBGaussians
from .Coordinates import DGBCartesians, DGBWatsonModes

__all__ = ["DGBWavefunctions", "DGBWavefunction"]

__reload_hook__ = ['.Components']

class DGBWavefunction(Wavefunction):
    def __init__(self, energy, data, gaussians:DGBGaussians = None, **opts):
        super().__init__(energy, data, **opts)
        if gaussians is None:
            gaussians = self.parent.gaussians
        self.gaussians = gaussians

    def get_dimension(self):
        return self.gaussians.coords.shape[-1]

    def plot(self,
             figure=None,
             domain=None,
             plot_centers=False,
             **opts
             ):

        centers = self.gaussians.coords.centers
        if domain is None:
            domain = Mesh(centers).bounding_box

        fig = super().plot(figure=figure, domain=domain, **opts)
        if isinstance(plot_centers, dict):
            plot_centers_opts = plot_centers.copy()
            plot_centers = True
        else:
            plot_centers_opts = {}
        if 'figure' not in plot_centers_opts:
            plot_centers_opts['figure'] = fig
        if plot_centers:
            if self.ndim == 1:
                plt.ScatterPlot(
                    centers[:, 0],
                    np.zeros(len(centers)),
                    **plot_centers_opts
                )
            elif self.ndim == 2:
                plt.ScatterPlot(
                    centers[:, 0],
                    centers[:, 1],
                    **plot_centers_opts
                )
            else:
                raise ValueError("can't plot centers for more than 2D data...?")
        return fig

    def to_cartesian_wavefunction(self):
        """
        Projects the wavefunction back to Cartesians
        :return:
        """
        if isinstance(self.gaussians.coords, DGBCartesians):
            return self

        new_gauss = self.gaussians.as_cartesians()
        return type(self)(
            self.energy,
            self.data,
            new_gauss,
            **self.opts
        )

    def plot_cartesians(self,
                        xyz_sel=None,
                        *,
                        atom_sel=None,
                        figure=None,
                        plot_centers=False,
                        atom_styles=None,
                        **plot_styles
                        ):

        if isinstance(self.gaussians.coords, DGBWatsonModes):
            return self.to_cartesian_wavefunction().plot_cartesians(
                xyz_sel=xyz_sel,
                atom_sel=atom_sel,
                figure=figure,
                plot_centers=plot_centers,
                atom_styles=atom_styles,
                **plot_styles
            )
        if not isinstance(self.gaussians.coords, DGBCartesians):
            raise ValueError("can't plot projections onto Cartesian coordinates for coords of type {}".format(
                type(self.gaussians.coords).__name__
            ))

        _, natoms, nxyz = self.gaussians.coords.cart_shape

        full_spec = np.arange(natoms*nxyz).reshape((natoms, nxyz))
        if (xyz_sel is None and nxyz > 2) or (xyz_sel is not None and len(xyz_sel) > 2):
            raise ValueError("can't plot 3D Gaussians")
        all_ats = np.arange(natoms)
        if atom_sel is None:
            atom_sel = all_ats

        if atom_styles is None:
            atom_styles = [{} for _ in atom_sel]

        if xyz_sel is not None:
            xyz_remv = np.setdiff1d(np.arange(nxyz), xyz_sel)
        else:
            xyz_remv = None
        for i,atom in enumerate(atom_sel):
            rem = np.setdiff1d(all_ats, [atom])
            subspec = full_spec[rem, :].flatten()
            if xyz_sel is not None:
                subspec = np.sort(
                    np.concatenate([subspec, full_spec[atom][xyz_remv]])
                )
            proj = self.marginalize_out(subspec)  # what we're projecting _out_

            ps = dict(plot_styles, **atom_styles[i])
            ps['plotter'] = ps.get(
                'plotter',
                plt.TriContourLinesPlot if proj.ndim == 2 else None
            )

            figure = proj.plot(
                figure=figure,
                plot_centers=plot_centers,
                # levels=np.linspace(-max_val, max_val, 16),
                # domain=[[-2, 2], [-2, .2]],
                # cmap='RdBu'
                **ps
            )

        return figure

    def evaluate(self, points):
        # raise NotImplementedError(...)
        points = np.asanyarray(points)
        reshape = None
        if points.ndim == 1:
            points = points[np.newaxis]
        elif points.ndim > 2:
            reshape = points.shape[:-1]
            points.reshape(-1, points.shape[-1])

        centers = self.gaussians.coords.centers
        # if self.mass_weighted:
        #     points = points * np.sqrt(self.masses[np.newaxis])
        #     centers = centers * np.sqrt(self.masses[np.newaxis])


        # actual basis evaluation
        c_disps = points[np.newaxis, :, :] - centers[:, np.newaxis, :]
        tfs = self.gaussians.transformations
        if tfs is not None:
            tfs, inv = tfs
            tfs = np.broadcast_to(tfs[:, np.newaxis, :, :], (len(tfs), len(points)) + tfs.shape[1:])
            c_disps = tfs.transpose((0, 1, 3, 2)) @ c_disps[:, :, :, np.newaxis]
            c_disps = c_disps.reshape(c_disps.shape[:-1])

        alphas = self.gaussians.alphas[:, np.newaxis, :]
        # if self.inds is not None:
        #     c_disps = c_disps[..., self.inds]
        #     alphas = alphas[..., self.inds]
        normas = np.prod((2 * alphas / np.pi) ** (1/4), axis=-1)
        exp_evals = np.prod(np.exp(-alphas * c_disps**2), axis=-1)

        momenta = self.gaussians.momenta
        if momenta is not None:
            points = points[np.newaxis, :, :, np.newaxis]
            if tfs is not None:
                points = tfs.transpose((0, 1, 3, 2)) @ points
            momenta = momenta[:, np.newaxis, np.newaxis, :]
            correlation = momenta @ points
            correlation = correlation.reshape(correlation.shape[:2])
            exp_evals *= np.cos(correlation)
            sum_term = np.sum(
                (momenta.reshape(alphas.shape)**2) / (2 * alphas),
                axis=-1
            )
            cos_corr = momenta @ centers[:, np.newaxis, :, np.newaxis]
            cos_corr = np.reshape(cos_corr, cos_corr.shape[:2])
            normas /= np.sqrt((1 + np.exp(-sum_term)*np.cos(cos_corr))/2)

        vals = np.dot(
            self.data,
            normas * exp_evals
        )
        if reshape is not None:
            vals = vals.reshape(reshape)
        return vals

    def marginalize_out(self, dofs, rescale=True) -> 'DGBWavefunction':
        """
        Computes the projection of the current wavefunction onto a set of degrees
        of freedom, returning a projected wave function object

        :return:
        :rtype: Wavefunction
        """

        scaling, subgaussians = self.gaussians.marginalize_out(dofs)

        return type(self)(
            self.energy,
            self.data * scaling,
            subgaussians
            )

class DGBWavefunctions(Wavefunctions):
    wavefunction_class = DGBWavefunction
    def __init__(self, energies=None, wavefunctions=None, hamiltonian=None,  **opts):
        super().__init__(energies=energies, wavefunctions=wavefunctions, hamiltonian=hamiltonian, **opts) # add all opts
        self._hamiltonian = hamiltonian
        self.gaussians = self.hamiltonian.gaussians
    def as_cartesian_wavefunction(self):
        """
        Projects the wavefunction back to Cartesians
        :return:
        """
        if isinstance(self.gaussians.coords, DGBCartesians):
            return self
        new_ham = self.hamiltonian.as_cartesian_dgb()
        new = type(self)(
            self.energies,
            self.wavefunctions,
            hamiltonian=new_ham
        )

        return new

        # raise Exception(
        #     self[0].evaluate(
        #         self.gaussians.coords.centers[:5]
        #     ),
        #     new[0].evaluate(
        #         new.gaussians.coords.centers[:5]
        #     )
        # )

    @property
    def hamiltonian(self):
        # if self._hamiltonian is None:
        #     from .DGB import DGB
        #     self._hamiltonian = DGB(
        #         self.centers,
        #         potential_function=None,
        #         alphas=self.alphas,
        #         transformations=self.transformations,
        #         min_singular_value=None,
        #         num_svd_vectors=None,
        #         clustering_radius=None
        #     )
        return self._hamiltonian
    def operator_representation(self,
                                op,
                                embed=True,
                                expansion_degree=None,
                                quadrature_degree=None,
                                expansion_type=None,
                                ):
        return self.hamiltonian.evaluate_multiplicative_operator(
            op,
            embed=embed,
            expansion_degree=expansion_degree,
            expansion_type=expansion_type,
            quadrature_degree=quadrature_degree
        )

    def expectation(self, op,
                    expansion_degree=None,
                    quadrature_degree=None,
                    expansion_type=None,
                    embed=True,
                    other=None
                    ):
        """Computes the expectation value of operator op over the wavefunction other and self

        :param other:
        :type other: Wavefunction | np.ndarray
        :param op:
        :type op:
        :return:
        :rtype:
        """

        if other is None:
            other = self

        if other.hamiltonian is not self.hamiltonian:
            raise ValueError("mismatch in DGBs between {} and {}".format(
                self, other
            ))

        op_mat = self.operator_representation(
            op,
            embed=embed,
            expansion_degree=expansion_degree,
            expansion_type=expansion_type,
            quadrature_degree=quadrature_degree
        )

        w1 = self.wavefunctions.T
        w2 = other.wavefunctions
        if op_mat.ndim > 2: # lots of shape fuckery to get broacasting right
            op_mat = np.moveaxis(np.moveaxis(op_mat, 0, -1), 0, -1)
            for _ in range(op_mat.ndim - 2):
                w1 = np.expand_dims(w1, 0)
                w2 = np.expand_dims(w2, 0)
            res = np.moveaxis(np.moveaxis(w1 @ op_mat @ w2, -1, 0), -1, 0)
        else:
            res = w1 @ op_mat @ w2

        return res

    def localize(self,
                 criterion,
                 which=None
                 ):
        """
        Find a transformation that maximally localizes the wavefunctions in the Boys' sense
        by minimizing <r^2> - <r>^2 over unitary transformations

        :param criterion:
        :param which:
        :return:
        """
        raise NotImplementedError('woof...')

        ndim = self.centers.shape[0]
        def r2_func(centers, deriv_order=2):
            c2 = centers**2

        r2 = self.operator_representation(

        )
        r = self.hamiltonian.S


    def __repr__(self):
        return "{}(num={}, DVR={})".format(
            type(self).__name__,
            len(self),
            self.hamiltonian
        )


'''
A package for doing generalized DVR in python.
Provides and extensible DVR framework with an easy-to-write structure.
'''

#TODO: migrate Mathematica structure -> more DVR classes, allow for a general direct-product DVR
#       provide hook-in for non-direct-product couplings, allow for coordinate-dependent mass

__all__= [ ]
from .DVR import *; from .DVR import __all__ as exposed
__all__ += exposed
from .BaseDVR import *; from .BaseDVR import __all__ as exposed
__all__ += exposed
from .ColbertMiller import *; from .ColbertMiller import __all__ as exposed
__all__ += exposed
from .DirectProduct import *; from .DirectProduct import __all__ as exposed
__all__ += exposed
from .Wavefunctions import *; from .Wavefunctions import __all__ as exposed
__all__ += exposed
from .Extensions import *; from .Extensions import __all__ as exposed
__all__ += exposed



import functools

import numpy as np
from ..VSCF import GridSCF
from McUtils.Scaffolding import ParameterManager

from .BaseDVR import BaseDVR
from .Wavefunctions import DVRWavefunctions
from .DirectProduct import DirectProductDVR
from .FiniteBasisDVR import WavefunctionBasisDVR

__all__ = [
    "SelfConsistentDVR",
    "PotentialOptimizedDVR"
]

class SCFWavefunctionGenerator:
    def __init__(self, dvr_1D:BaseDVR):
        self.dvr = dvr_1D
        self.prev = None
    def __call__(self, pot, **kwargs):
        if self.prev is None:
            res = self.prev = self.dvr.run(potential_values=pot)
        else:
            res = self.dvr.run(
                potential_values=pot,
                grid=self.prev.grid,
                kinetic_energy=self.prev.kinetic_energy
            )
        return res.wavefunctions

class SelfConsistentDVR(GridSCF):
    def __init__(self, base_dvr:"DirectProductDVR", **opts):
        props = ParameterManager(**opts)
        self.base_dvr = base_dvr
        grid, pe = self.create_grid_vals()
        generators = self.create_solvers(grid, pe)
        super().__init__(grid, pe, generators, **props.filter(GridSCF))
    # def initialize(self):
    #     d = super().initialize()
    #     for w in d.wavefunctions:
    #         w[:1].plot()
    #     import McUtils.Plots as plt
    #     plt.DensityPlot(*self.grid.transpose(2, 0, 1), self.vals, plot_style=dict(vmin=0, vmax=1)).show()
    #     return d
    def create_grid_vals(self):
        pot_data = self.base_dvr.run(result='potential_energy')
        grid = pot_data.grid
        pe = pot_data.potential_energy
        if hasattr(pe, 'diagonal'):
            pe = pe.diagonal()
        else:
            pe = np.diag(pe)
        pe = pe.reshape(grid.shape[:-1])
        return grid, pe
    def create_solvers(self, grid, pe):
        # rebind the
        init = self.get_initial_point(pe)
        gp = grid[init]
        for i, d in enumerate(self.base_dvr.dvrs):
            g = d.opts.get('g', None)
            if g is not None:
                try:
                    iter(g)
                except TypeError:
                    g = self._wrap_g(g, gp, i)
                else:
                    g = [
                        [self._wrap_g(el, gp, i) for el in _]
                        for _ in g
                    ]
                d.opts['g'] = g

                gd = d.opts.get('g_deriv', None)
                if gd is not None:
                    try:
                        iter(gd)
                    except TypeError:
                        gd = self._wrap_g(gd, gp, i)
                    else:
                        gd = [self._wrap_g(el, gp, i) for el in gd]
                    d.opts['g_deriv'] = gd

        generators = [
            SCFWavefunctionGenerator(d)
            for d in self.base_dvr.dvrs
        ]

        return generators
    @staticmethod
    def _wrap_g(g, gp, i):
        if not callable(g):
            return g
        @functools.wraps(g)
        def eval_g(grid, g=g):
            full_ars = [np.full_like(grid, x) for x in gp]
            full_ars[i] = grid
            grid = np.concatenate(
                [np.expand_dims(a, -1) for a in full_ars],
                axis=1
            )
            return g(grid)
        return eval_g

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.base_dvr
        )

class PotentialOptimizedDVR(DirectProductDVR):
    def __init__(self,
                 wfns_1D:'Iterable[DVRWavefunctions]',
                 **base_opts
                 ):
        # base_opts = {k:base_opts[k] for k in base_opts.keys() - {'mass', 'g', 'g_deriv', 'include_kinetic_coupling'}}
        super().__init__(
            [WavefunctionBasisDVR(w) for w in wfns_1D],
            **base_opts
        )

    @classmethod
    def from_minimum(cls, base_dvr:"DirectProductDVR|SelfConsistentDVR", **opts):
        if not isinstance(base_dvr, SelfConsistentDVR):
            base_dvr = SelfConsistentDVR(base_dvr)
        wfns = base_dvr.initialize().wavefunctions
        return cls(
            wfns,
            **dict(
                base_dvr.base_dvr.opts,
                **opts
            )
        )

    @classmethod
    def from_scf(cls, scf_dvr:"DirectProductDVR|SelfConsistentDVR", wfns=None, **opts):
        if not isinstance(scf_dvr, SelfConsistentDVR):
            scf_dvr = SelfConsistentDVR(scf_dvr)
        if wfns is None:
            wfns = scf_dvr.run().wavefunctions
        return cls(
            wfns,
            **dict(
                scf_dvr.base_dvr.opts,
                **opts
            )
        )


"""
Provides concrete tools for dealing with two of the most useful types of surfaces we have
"""

import numpy as np
from collections import namedtuple
from McUtils.GaussianInterface import GaussianLogReader, GaussianFChkReader
from McUtils.Zachary import Surface, MultiSurface, InterpolatedSurface, TaylorSeriesSurface

__all__=[
    "DipoleSurface",
    "PotentialSurface"
]

class DipoleSurface(MultiSurface):
    """
    Provides a unified interface to working with dipole surfaces.
    Currently basically no fancier than a regular surface (although with convenient loading functions), but dipole-specific
    stuff could come
    """
    def __init__(self, mu_x, mu_y, mu_z):
        """

        :param mu_x: X-component of dipole moment
        :type mu_x: Surface
        :param mu_y: Y-component of dipole moment
        :type mu_y: Surface
        :param mu_z: Z-component of dipole moment
        :type mu_z: Surface
        """
        if isinstance(mu_x.base, TaylorSeriesSurface):
            self.mode = "taylor"
        else:
            self.mode = "interp"
        super().__init__(
            mu_x,
            mu_y,
            mu_z
        )

    @staticmethod
    def get_log_values(log_file, keys=("StandardCartesianCoordinates", "DipoleMoments")):

        with GaussianLogReader(log_file) as parser:
            parse_data = parser.parse(keys)

        carts = parse_data[keys[0]][1]
        dipoles = np.array(parse_data[keys[1]])

        return namedtuple('dipole_log_values', ['cartesians', 'dipoles'])(carts, dipoles)
    @classmethod
    def from_log_file(cls, log_file, coord_transf, keys=("StandardCartesianCoordinates", "DipoleMoments"), tol = .001, **opts):
        """
        Loads dipoles from a Gaussian log file and builds a dipole surface by interpolating.
        Obviously this only really works if we have a subset of "scan" coordinates, so at this stage the user is obligated
        to furnish a function that'll take a set of Cartesian coordinates and convert them to "scan" coordinates.
        Coordinerds can be helpful with this, as it provides a convenient syntax for Cartesian <-> ZMatrix conversions

        :param log_file: a Gaussian log file to pull from
        :type log_file: str
        :return:
        :rtype:
        """

        carts, dipoles = cls.get_log_values(log_file, keys=keys)

        scan_coords = coord_transf(carts)
        if len(dipoles) != len(scan_coords):
            raise ValueError(
                "mismatch between number of dipoles ({}) and number of coordinates ({})".format(
                    len(dipoles),
                    len(scan_coords)
                )
            )

        if scan_coords.ndim == 1:
            scan_sort = np.argsort(scan_coords)
        else:
            scan_sort = np.lexsort(tuple(reversed(tuple(scan_coords.T))))
        scan_coords = scan_coords[scan_sort]
        dipoles = dipoles[scan_sort]

        # this is particularly relevant for optimization scans...but we pull all the "unique" indices
        # then we pull the indices right _before_ each unique one since that's the final one in the block of "uniques"
        # finally we do a "roll" to make sure the order from the sort is preserved
        tol_coords = np.floor(scan_coords/tol)
        if tol_coords.ndim == 1:
            diffs = np.diff(tol_coords)
        else:
            diffs = np.sum(abs(np.diff(tol_coords, axis=0)), axis=1)
        inds = np.where(diffs != 0)[0]
        inds = np.concatenate((inds, [len(inds)]))
        scan_coords = scan_coords[inds]
        dipoles = dipoles[inds]

        dipoles = list(np.transpose(dipoles))

        return DipoleSurface(*(
            Surface(
                ((scan_coords, d), opts),
                base = InterpolatedSurface,
                dipole_component = "x" if i == 0 else "y" if i == 1 else "z"
            ) for i,d in enumerate(dipoles)
        ))

    @staticmethod
    def get_fchk_values(fchk_file):

        with GaussianFChkReader(fchk_file) as parser:
            parse_data = parser.parse(["Coordinates", "Dipole Moment", "Dipole Derivatives"])

        center = parse_data["Coordinates"]
        const_dipole = parse_data["Dipole Moment"]
        derivs = parse_data["Dipole Derivatives"]
        derivs = np.reshape(derivs, (int(len(derivs) / 3), 3))

        return namedtuple('dipole_fchk_values', ['center', 'const', 'derivs'])(center, const_dipole, derivs)
    @classmethod
    def from_fchk_file(cls, fchk_file, **opts):
        """
        Loads dipoles from a Gaussian formatted checkpoint file and builds a dipole surface via a linear approximation

        :param fchk_file: a Gaussian fchk file to pull from
        :type log_file: str
        :return:
        :rtype:
        """

        center, const_dipole, derivs = cls.get_fchk_values(fchk_file)
        derivs = list(np.transpose(derivs))

        opts['center'] = center.flatten()
        surfs = [None]*3
        for i, d in enumerate(zip(derivs, list(const_dipole))):
            d, r = d
            opts = opts.copy()
            opts["ref"] = r
            surfs[i] = Surface(
                ((d,), opts),
                base = TaylorSeriesSurface,
                dipole_component="x" if i == 0 else "y" if i == 1 else "z"
            )

        return cls(*surfs)

    def __call__(self, gridpoints, **opts):
        """
        Explicitly overrides the Surface-level evaluation because we know the Taylor surface needs us to flatten our gridpoints

        :param gridpoints:
        :type gridpoints:
        :param opts:
        :type opts:
        :return:
        :rtype:
        """

        gps = np.asarray(gridpoints)
        if self.mode == "taylor":
            if gps.ndim == 2:
                gps = gps.flatten()
            elif gps.ndim > 2:
                gps = np.reshape(gps, gps.shape[:-2] + (np.prod(gps.shape[-2:]),))

        return super().__call__(gps, **opts)

class PotentialSurface(Surface):
    """
    A potential surface structure to go along with the DipoleSurface.
    Provides convenient access to potential data + a unified interface to things like energy minimization
    """

    @staticmethod
    def get_log_values(log_file, keys=("StandardCartesianCoordinates", "ScanEnergies")):

        with GaussianLogReader(log_file) as parser:
            parse_data = parser.parse(keys)

        # Need to be smarter about this. At some point we might be able to infer what type of log file we have...
        coord_key = keys[0]
        coords = parse_data[coord_key][1]
        eng_key = keys[1]
        if eng_key == "ScanEnergies":
            energies = np.array(parse_data[eng_key].energies[:, -1])
        else:
            raise Exception("Haven't dealt with scan types beyond rigid ones...")

        return namedtuple('potential_log_values', ['coords', 'energies'])(coords, energies)
    @classmethod
    def from_log_file(cls, log_file, coord_transf, keys=("StandardCartesianCoordinates", "ScanEnergies"), tol = .001, **opts):
        """
        Loads dipoles from a Gaussian log file and builds a potential surface by interpolating.
        Obviously this only really works if we have a subset of "scan" coordinates, so at this stage the user is obligated
        to furnish a function that'll take a set of Cartesian coordinates and convert them to "scan" coordinates.
        Coordinerds can be helpful with this, as it provides a convenient syntax for Cartesian <-> ZMatrix conversions.

        :param log_file: a Gaussian log file to pull from
        :type log_file: str
        :return:
        :rtype:
        """

        dat = cls.get_log_values(log_file, keys=keys)
        carts = dat.coords
        pots = dat.energies

        # raise Exception(carts, pots)
        scan_coords = coord_transf(carts)
        if len(pots) != len(scan_coords):
            raise ValueError(
                "mismatch between number of potential values ({}) and number of coordinates ({})".format(
                    len(pots),
                    len(scan_coords)
                )
            )

        if scan_coords.ndim == 1:
            scan_sort = np.argsort(scan_coords)
        else:
            scan_sort = np.lexsort(tuple(reversed(tuple(scan_coords.T))))
        scan_coords = scan_coords[scan_sort]
        pots = pots[scan_sort]

        # this is particularly relevant for optimization scans...but we pull all the "unique" indices
        # then we pull the indices right _before_ each unique one since that's the final one in the block of "uniques"
        # finally we do a "roll" to make sure the order from the sort is preserved
        tol_coords = np.floor(scan_coords/tol)
        if tol_coords.ndim == 1:
            diffs = np.diff(tol_coords)
        else:
            diffs = np.sum(abs(np.diff(tol_coords, axis=0)), axis=1)
        inds = np.where(diffs != 0)[0]
        inds = np.concatenate((inds, [len(inds)]))
        scan_coords = scan_coords[inds].squeeze()
        pots = pots[inds]

        return cls(
                ((scan_coords, pots), opts),
                base=InterpolatedSurface
        )

    @staticmethod
    def get_fchk_values(fchk_file):
        # TODO: I know I probably didn't do this right but I'm just getting a thing out for now
        with GaussianFChkReader(fchk_file) as parser:
            parse_data = parser.parse(["Coordinates", "Total Energy", "Gradient", "ForceConstants", "ForceDerivatives"])

        center = parse_data["Coordinates"]
        eng = parse_data["Total Energy"]
        derivs = [parse_data['Gradient'], parse_data["ForceConstants"], parse_data["ForceDerivatives"]]

        return namedtuple('potential_fchk_values', ['center', 'energy', 'derivs'])(
            center, eng, derivs
        )
    @classmethod
    def from_fchk_file(cls, fchk_file, **opts):
        """
        Loads potential from a Gaussian formatted checkpoint file and builds a potential surface via a quartic approximation

        :param fchk_file: a Gaussian fchk file to pull from
        :type log_file: str
        :return:
        :rtype:
        """

        center, energy, derivs = cls.get_fchk_values(fchk_file)

        return cls((derivs, dict(ref=energy, center=center.flatten())), base=TaylorSeriesSurface, **opts)


"""
Provides a general class of localized modes using a potentially non-linear
transformation of normal modes
"""

import numpy as np, scipy
import McUtils.Numputils as nput

from .MixtureModes import *
from .NormalModes import *

__all__ = ['LocalizedModes']
class LocalizedModes(MixtureModes):

    def __init__(self,
                 normal_modes: NormalModes,
                 transformation,
                 inverse=None,
                 origin=None,
                 masses=None,
                 freqs=None,
                 mass_weighted=None,
                 frequency_scaled=None,
                 **etc
                 ):
        mat = normal_modes.matrix @ transformation
        if inverse is None:
            inverse = transformation.T
        self.localizing_transformation = (transformation, inverse)
        self.base_modes = normal_modes
        inverse = inverse @ normal_modes.inverse
        if origin is None:
            origin = normal_modes.origin
        if masses is None:
            masses = normal_modes.masses
        if freqs is None:
            freqs = normal_modes.freqs  # this preserves information, but could be confusing...
        super().__init__(
            normal_modes.basis,
            mat,
            inverse=inverse,
            origin=origin,
            masses=masses,
            freqs=freqs,
            mass_weighted=normal_modes.mass_weighted,
            frequency_scaled=normal_modes.frequency_scaled,
            **etc
        )

    @property
    def mass_weighted(self):
        return self.base_modes.mass_weighted
    @mass_weighted.setter
    def mass_weighted(self, new):
        if (
                new and not self.base_modes.mass_weighted
                or not new and self.base_modes.mass_weighted
        ):
            raise ValueError("can't set `mass_weighted` directly")
    @property
    def frequency_scaled(self):
        return self.base_modes.frequency_scaled
    @frequency_scaled.setter
    def frequency_scaled(self, new):
        if (
                new and not self.base_modes.frequency_scaled
                or not new and self.base_modes.frequency_scaled
        ):
            raise ValueError("can't set `frequency_scaled` directly")
    @property
    def g_matrix(self):
        return self.matrix.T @ self.base_modes.g_matrix @ self.matrix
    @g_matrix.setter
    def g_matrix(self, g):
        ...

    def modify(self,
               base_modes=None,
               *,
               transformation=None,
               freqs=None,
               origin=None,
               masses=None,
               inverse=None,
               name=None,
               mass_weighted=None,
               frequency_scaled=None,
               g_matrix=None
               ):
        return type(self)(
            self.base_modes if base_modes is None else base_modes,
            self.localizing_transformation[0] if transformation is None else transformation,
            inverse=self.localizing_transformation[1] if inverse is None else inverse,
            origin=origin,
            masses=masses,
            freqs=freqs,
            g_matrix=g_matrix,
            name=self.name if name is None else name
        )

    def make_mass_weighted(self, **kwargs):
        return self.modify(self.base_modes.make_mass_weighted(**kwargs))
    def remove_mass_weighting(self, **kwargs):
        return self.modify(self.base_modes.remove_mass_weighting(**kwargs))
    def make_frequency_scaled(self, **kwargs):
        return self.modify(self.base_modes.make_frequency_scaled(**kwargs))
    def remove_frequency_scaling(self, **kwargs):
        return self.modify(self.base_modes.remove_frequency_scaling(**kwargs))

    @property
    def local_freqs(self):
        return np.diag(self.local_hessian)

    @property
    def local_hessian(self):
        tf, inv = self.localizing_transformation
        f = inv @ np.diag(self.freqs ** 2) @ inv.T
        g = self.g_matrix
        a = np.diag(np.power(np.diag(g) / np.diag(f), 1/4))
        return a @ f @ a



import numpy as np, abc
from .StateSpaces import BasisStateSpace, BasisMultiStateSpace, SelectionRuleStateSpace

__all__ = [
    "BasisStateSpaceFilter"
]

class Postfilter(metaclass=abc.ABCMeta):
    def __init__(self, input_space=None):
        self.input_space = input_space
    @abc.abstractmethod
    def apply(self, state_space:BasisStateSpace)->BasisStateSpace:
        pass
    @classmethod
    def get_spec_map(cls):
        return {
            'test':ExclusionTestFilter,
            'max_quanta':MaxQuantaFilter,
            'intersected_space':IntersectionSpaceFilter,
            'excluded_space':DifferenceSpaceFilter,
            'excluded_transitions':ExcludedTransitionFilter,
        }
    @classmethod
    def construct(cls, input_space, spec):
        if isinstance(spec, BasisStateSpace):
            spec = {'intersected_space':spec}
        for k,c in cls.get_spec_map().items():
            if k in spec:
                return c(input_space=input_space, **spec)
        else:
            raise ValueError("Don't know how to build filter from spec {}".format(spec))

class ExclusionTestFilter(Postfilter):
    def __init__(self, test=None, **opts):
        super().__init__(**opts)
        self.test = test
    def apply(self, state_space:BasisStateSpace) ->BasisStateSpace:
        vals = self.test(state_space)
        where = np.where(vals)
        if len(where) > 0:
            where = where[0]
        return state_space.take_subspace(where)
class MaxQuantaFilter(ExclusionTestFilter):
    def __init__(self, max_quanta=None, input_space:BasisStateSpace=None, **opts):
        if isinstance(max_quanta, (int, np.integer)):
            max_quanta = [max_quanta] * input_space.ndim
        self.max_quanta = np.asanyarray(max_quanta)
        super().__init__(self.check_mode_quanta, input_space=input_space, **opts)
    def check_mode_quanta(self, state_space):
        if len(state_space) == 0:
            return np.array([], dtype=bool)
        exc = state_space.excitations
        qs = self.max_quanta.copy()
        qs[qs < 0] = np.max(exc)+1
        return np.all(exc <= qs[np.newaxis], axis=1)
class IntersectionSpaceFilter(Postfilter):
    def __init__(self, intersected_space=None, **opts):
        super().__init__(**opts)
        self.space = intersected_space
    def apply(self, state_space:BasisStateSpace) ->BasisStateSpace:
        return state_space.intersection(self.space)
class DifferenceSpaceFilter(Postfilter):
    def __init__(self, excluded_space=None, **opts):
        super().__init__(**opts)
        self.space = excluded_space
    def apply(self, state_space:BasisStateSpace) ->BasisStateSpace:
        return state_space.difference(self.space)
class ExcludedTransitionFilter(DifferenceSpaceFilter):
    def __init__(self, excluded_transitions=None, input_space=None, **opts):
        bad_transitions = np.asanyarray(excluded_transitions)
        perms = (
                input_space.excitations[:, np.newaxis, :]
                + bad_transitions[np.newaxis, :, :]
        ).reshape(-1, bad_transitions.shape[-1])
        perms = perms[np.all(perms > 0, axis=1)]
        excluded_space = BasisStateSpace(input_space.basis, perms)
        super().__init__(excluded_space=excluded_space, input_space=input_space, **opts)
    def apply(self, state_space:BasisStateSpace) ->BasisStateSpace:
        return state_space.difference(self.space)

class BasisStateSpaceFilter:
    """
    Provides an easier constructor for the VPT state space filters
    """

    def __init__(self, input_space, prefilters, postfilters):
        """
        :param input_space:
        :type input_space: BasisStateSpace
        :param prefilters:
        :type prefilters:
        :param postfilters:
        :type postfilters:
        """
        self.input_space = input_space
        self._prefilters = None
        self._raw_prefilters = prefilters
        self._postfilters = None
        self._raw_postfilters = postfilters

    @classmethod
    def from_data(cls, input_space, data):
        """
        Works to canonicalize inputs and initialize appropriately from there

        :param data:
        :type data:
        :return:
        :rtype:
        """

        if data is None:
            return None
        elif isinstance(data, dict):
            if not all(isinstance(v, cls) for v in data.values()):
                return cls._from_old_style_rules(input_space, data)
            else:
                return data
        else:
            return cls.from_rules(input_space, *data)

    @classmethod
    def _from_old_style_rules(cls, input_space, rules):
        """
        Builds a set of filter spaces from a dict keyed by Hamiltonian/wfn pairs
        and giving prefilters/postfilters

        :param rules:
        :type rules:
        :return:
        :rtype:
        """
        new_rules = []
        for k,v in rules.items():
            if not isinstance(v, dict):
                v = {'prefilters':v}
            new_rules.append(dict({'order':k[0]+k[1], 'expansion_term':k[0]}, **v))

        return cls.from_rules(input_space, *new_rules)


    @classmethod
    def from_rules(cls, input_space, *rules):
        """
        Builds a set of filter spaces from dicts of rules

        :param rules:
        :type rules:
        :return:
        :rtype:
        """

        filters = {}
        for r in rules:
            if not isinstance(r, dict):
                raise ValueError("filter rule must be a dict with `order`, `expansion_term`"
                                 ", and one or both of `prefilters` and `postfilter` for keys")
            for k in ["order", "expansion_term"]:
                if k not in r:
                    raise ValueError("filter rule needs `order`, `expansion_term`"
                                     ", and one or both of `prefilters` and `postfilter` for keys")
            if 'prefilters' not in r and 'postfilter' not in r:
                raise ValueError("filter rule needs one or both of `prefilters` and `postfilter` for keys")

            h_term = r['expansion_term']
            y_term = r['order'] - h_term

            filters[(h_term, y_term)] = cls(
                input_space,
                r['prefilters'] if 'prefilters' in r else None,
                r['postfilters'] if 'postfilters' in r else None
            )

        return filters

    @property
    def prefilters(self):
        """

        :return:
        :rtype:
        """
        if self._prefilters is None:
            self._prefilters = self.canonicalize_prefilters(self.input_space.basis,
                                                            self._raw_prefilters)
        return self._prefilters

    @property
    def postfilters(self):
        """

        :return:
        :rtype:
        """
        if self._postfilters is None:
            self._postfilters = self.canonicalize_postfilters(self.input_space,
                                                            self._raw_postfilters)
        return self._postfilters
    @classmethod
    def canonicalize_postfilters(self, input_space, filters):
        if filters is None:
            return None
        else:
            if hasattr(filters, 'items') or isinstance(filters, BasisStateSpace):
                filters = [filters]
            return [
                Postfilter.construct(input_space, f) for f in filters
            ]


    def _could_be_a_space(self, test): # nasty checking code that I don't want to redupe all the time
        if isinstance(test, (BasisStateSpace, BasisMultiStateSpace)):
            return True
        else:
            try:
                lt = len(test)
            except TypeError:
                return False
            else:
                if lt == 0:
                    return True

            try:
                if lt > 0 and isinstance(test[0], (int, np.integer)):
                    return True
                else:
                    lt = len(test[0])
                    if lt > 0 and isinstance(test[0][0], (int, np.integer)):
                        return True
                    else:
                        return False
            except (TypeError, IndexError):
                return False

    def _could_be_rules(self, test):
        # selection rule options
        if test is None:
            return True

        try:
            lt = len(test)
        except TypeError:
            return False

        if lt == 0:
            return True

        try:
            lt = len(test[0])
            res = (
                    lt == 0
                    or isinstance(test[0][0], (int, np.integer))
            )
        except TypeError:
            return False

        return res

    def _could_be_a_prefilter(self, test):
        return (
            test is None
            or self._could_be_a_space(test)
            or (
                    len(test) == 2
                    and (test[0] is None or self._could_be_a_space(test[0]))
                    and (self._could_be_rules(test[1]))
            )
        )

    def canonicalize_prefilters(self, basis, prefilters):
        """
        Puts the prefilters in canonical form...

        :param basis:
        :type basis:
        :param prefilters:
        :type prefilters:
        :return:
        :rtype:
        """

        if prefilters is None:
            return None

        if self._could_be_a_prefilter(prefilters):
            prefilters = (prefilters,)

        final_filters = []
        for n, filter_space in enumerate(prefilters):
            if filter_space is None or self._could_be_a_space(filter_space):
                filter_rules = None
            else:
                filter_space, filter_rules = filter_space

            if filter_space is not None:
                if isinstance(filter_space, (int, np.integer)):
                    filter_space = BasisStateSpace.from_quanta(
                        self.input_space.basis,
                        (filter_space,)
                    )
                elif isinstance(filter_space, (list, tuple)) and all(
                        isinstance(f, (int, np.integer)) for f in filter_space
                ):
                    filter_space = BasisStateSpace.from_quanta(
                        self.input_space.basis,
                        filter_space
                    )
                if not isinstance(filter_space, (BasisStateSpace, BasisMultiStateSpace)):
                    filter_space = BasisStateSpace(basis, filter_space)
                if isinstance(filter_space, SelectionRuleStateSpace):
                    filter_space = filter_space.to_single().take_unique()

            final_filters.append([filter_space, filter_rules])

        return final_filters

    @classmethod
    def from_property_rules(cls,
                            initial_space,
                            target_space,
                            perturbation_rules,
                            property_rules,
                            order=2,
                            postfilters=None
                            , **opts
                            ):
        """
        :param initial_space:
        :type initial_space:
        :param target_space:
        :type target_space:
        :param perturbation_rules:
        :type perturbation_rules:
        :param property_rules:
        :type property_rules:
        :param order:
        :type order:
        :return:
        :rtype:
        """
        int_rules = cls._create_prop_rules_tree(
            initial_space,
            target_space,
            perturbation_rules,
            property_rules,
            order=order,
            **opts
        )

        if isinstance(initial_space, BasisStateSpace):
            basis = initial_space.basis
            ndim = initial_space.ndim
        else:
            from .HarmonicOscillator import HarmonicOscillatorProductBasis
            ndim = len(initial_space[0])
            basis = HarmonicOscillatorProductBasis(ndim)

        if postfilters is None:
            postfilters = {}
            single_post = False
        else:
            single_post = not hasattr(postfilters, 'items')
            if not single_post:
                single_post = not any(
                    len(k) == 2 and isinstance(k[0], (int, np.integer)) and isinstance(k[1], (int, np.integer))
                    for k in postfilters
                )

        full_rules = {
            k:{
                'prefilters':tuple((BasisStateSpace.from_quanta(basis, [i]), r) for i,r in v.items()),
                'postfilters':(postfilters if single_post else (None if k not in postfilters else postfilters[k]))
            } for k,v in int_rules.items()
        }

        return full_rules

    @classmethod
    def _build_nquanta_trees(cls, initial_space, perturbation_rules, order):
        """
        Build the tree of paths from the initial/target spaces to the
        final numbers of quanta based on the perturbation rules

        :param initial_space:
        :type initial_space:
        :param target_space:
        :type target_space:
        :param perturbation_rules:
        :type perturbation_rules:
        :param order:
        :type order:
        :return:
        :rtype:
        """

        if isinstance(initial_space, BasisStateSpace):
            initial_space = initial_space.excitations
        init_nquanta = np.unique(np.sum(initial_space, axis=1))
        pert_rules_nquanta = [np.unique([sum(x) for x in p]) for p in perturbation_rules]
        init_places = [{k: [()] for k in init_nquanta}]
        for o in range(order):
            # we determine what rules apply at each order
            # and where this can take the initial and final spaces
            r1_dict = {}
            for j,r1 in enumerate(tuple(init_places)):
                p = pert_rules_nquanta[o-j]
                for r,v in r1.items():
                    for k in p:
                        key = (1+o-j, k)
                        n = r + k
                        if n >= 0:
                            if n not in r1_dict:
                                r1_dict[n] = [t+(key,) for t in v]
                            else:
                                r1_dict[n].extend(t+(key,) for t in v)
            init_places.append(r1_dict)

        return init_nquanta, init_places

    @classmethod
    def _create_prop_rules_tree(cls,
                                initial_space,
                                target_space,
                                perturbation_rules,
                                property_rules,
                                order=2,
                                eliminate_negatives=True,
                                subspace=True,
                                check_subspaces=True
                                ):

        # we really want to do this one at a time for every initial quanta
        # and target quanta pair

        init_nquanta, init_places = cls._build_nquanta_trees(initial_space, perturbation_rules, order)
        targ_nquanta, targ_places = cls._build_nquanta_trees(target_space, perturbation_rules, order)
        # print(init_places)

        # with those sets built we can now figure out which property rules can connect
        # the terms at which orders
        rules = []
        def check_zero_rule(x):
            return (
                sum(y[1] for y in x) == 0 # doesn't change number of quanta
                # and # and the term will be iterated enough this rule can apply
            )
        for k in range(order+1):
            for p1, v1 in init_places[k].items():
                clean_v = [x for x in v1 if check_zero_rule(x)]
                if len(clean_v)>0:
                    rules.append(
                        ((p1, clean_v), None, (None, None))
                    )
            for p2, v2 in targ_places[k].items():
                clean_v = [x for x in v2 if check_zero_rule(x)]
                if len(clean_v)>0:
                    rules.append(
                        ((None, None), None, (p2, clean_v))
                    )
            if property_rules is not None:
                prop_rules_nquanta = [np.unique([sum(x) for x in p]) for p in property_rules]
                for i in range(k+1):
                    for j in range(k-i+1):
                        matches = [
                            ((p1, v1), x, (p2, v2))
                            for p1,v1 in init_places[i].items()
                            for p2,v2 in targ_places[j].items()
                            for x in prop_rules_nquanta[k-i-j]
                            if p2+x==p1 or p1+x==p2
                        ]
                        rules.extend(matches)

        # now we need to take these matches and use them to determine which terms are truly needed
        # for both the initial and target spaces
        keys = {}
        for (n,v), _, _ in rules:
            if n is not None:
                if n not in keys:
                    keys[n] = set(v)
                else:
                    keys[n].update(v)
        for _, _, (n,v) in rules:
            if n is not None:
                if n not in keys:
                    keys[n] = set(v)
                else:
                    keys[n].update(v)
        full_rules = cls._construct_rule_table_from_paths(perturbation_rules, keys)

        if eliminate_negatives:
            full_rules = {
                # we can eliminate rules that would require dropping below zero
                ord:{
                    k:[r for r in rule_vec if -sum(x for x in r if x < 0) <= k]
                    for k,rule_vec in rules.items()
                }
                for ord,rules in full_rules.items()
            }

        if isinstance(target_space, BasisStateSpace):
            target_space = target_space.excitations
        space_counts = dict(zip(*np.unique(np.sum(target_space, axis=1), return_counts=True)))
        quants = list(space_counts.keys())
        if check_subspaces:
            is_subspace = {
                k:space_counts[k] < v
                for k,v in zip(
                    quants,
                    BasisStateSpace.num_states_with_quanta(quants, len(target_space[0]))
                )
            }
        else:
            is_subspace = {}

        tree_ordering = list(sorted(full_rules.keys()))
        tree_groups = [[full_rules[x] for x in tree_ordering if x[0] == i+1] for i in range(order)]
        new_rules = {
            k:cls._prune_transition_tree(
                tree_groups[k[0]-1][:k[1]],
                t,
                is_subspace
            )
            for k,t in full_rules.items()
        }
        if property_rules is None:
            new_rules[(order, 0)] = {k:((),) for k in new_rules[(order, 0)]}
        # print(new_rules, order)

        # # Do one final prune based on rules
        # # this is necessary because when working with
        # # subspaces we can't make use of symmetry as effectively
        # # as we otherwise could
        # if property_rules is not None:
        #     prop_rules_nquanta = [np.unique([sum(x) for x in p]) for p in property_rules]
        #     for ord in range(order+1):
        #         for k in range(ord+1):
        #             for j in range(ord+1 - k): # all pairs of rules
        #                a = ord - k - j
        #                start_quanta = [
        #
        #                ]
        #                possible_qfs = [
        #                    [qi + p for qi in start_quanta for p in dqs if qi + p >= 0]
        #                    for dqs in prop_rules_nquanta
        #                ]
        #                for init_ham in range(1, k+1):
        #                    for final_ham in range()
        #                # for qi in init_nquanta:
        #                #     for qf in targ_nquanta:
        #                #         for k in
        #
        #
        #                print(k, a, j, prop_rules_nquanta[a])

        if not subspace: # TODO: infer subspace from number of states provided...
            # we're not transforming full spaces (index by quanta) and so we need to
            # include back in at higher order some of the transformations from lower order
            # that we excluded via symmetry
            for o in range(1, order+1):
                for j in range(1, order+1 - o): # all the higher order terms
                    prev = (o, j-1)
                    cur = (o, j)
                    new_terms = {}
                    max_quant = max(sum(t) + oo for oo,v in new_rules[prev].items() for t in v)
                    # print(max_quant, cur, prev)
                    for k, v in new_rules[cur].items():
                        base_rules = set(v)
                        added_terms = [
                            t for t in
                                new_rules[prev].get(k, ())
                            if 0 <= sum(t) + k and sum(t) + k <= max_quant
                        ]
                        # print(k, added_terms)
                        base_rules.update(added_terms)
                        new_terms[k] = tuple(base_rules)
                    new_rules[cur] = new_terms

        """
        {3: ((-2, 1), (-1,), (1, -1, -1)), 1: ((1, 1, -1), (1,), (2, -1)), 0: ((1,),)}
        {3: ((-2, 1), (-1,), (1, -1, -1)), 1: ((1, 1, -1), (1,), (2, -1)), 4: ((-1, -1, -1), (-3,), (-2, -1)), 0: ((1,),), 2: ((-2, 1), (-1,), (1, -1, -1))}
"""
        # del new_rules[(1, 1)][2]
        # del new_rules[(1, 1)][4]
        # new_rules[(1, 1)][1] = ((1, 1, -1), (1,), (2, -1), (3,), (2, 1))

        # raise Exception(new_rules[(1, 1)])

        return new_rules

    @classmethod
    def _prune_full_rules(self, init_space, targ_space, full_rules, property_rules, order):
        # we now check which rules truly can get us from the init space
        # to the targ space in under the target order
        raise NotImplementedError("???")
        for key, states in full_rules.items():
            ...


    @staticmethod
    def _construct_rule_table_from_paths(perturbation_rules, keys):
        init_full_rules = {}
        for t,r in keys.items():
            for k in r:
                cur_d = t - sum(d for h,d in k)  # this will track which numbers of quanta to filter on
                for n, (h, d) in enumerate(k):
                    # h is which Hamiltonian term we're working with
                    # and d is which displacement is allowed
                    key = (h, n)
                    subrules = set(s for s in perturbation_rules[h - 1] if sum(s) == d)
                    if key not in init_full_rules:
                        init_full_rules[key] = {cur_d:subrules}
                    elif cur_d not in init_full_rules[key]:
                        init_full_rules[key][cur_d] = subrules
                    else:
                        init_full_rules[key][cur_d].update(subrules)
                    cur_d += d
                # print(init_full_rules[key])

        return init_full_rules

    @staticmethod
    def _prune_transition_tree(prev_trees, ttree, subspace_specs):
        # remove terms by symmetry
        new_tree = {}
        for k,paths in ttree.items():
            # we go through the proposed changes for the current
            # tree and check if any of them have a symmetric
            # pair starting from a state with fewer quanta
            #
            # We also need to be careful about subspaces, in particular
            # if we are transforming from a subspace, we can miss terms
            # notably when a state is accessed by an intermediate
            subpaths = {}
            for v in paths:
                shift = sum(v)
                new = shift + k
                duped = False
                if new < k:
                    w = tuple(sorted((-x for x in v), key=lambda x:-100*abs(x)-x))
                    for t in prev_trees:
                        if new in t:
                            if w in t[new]:
                                if subspace_specs.get(new, False):
                                    if new not in subpaths:
                                        subpaths[new] = []
                                    subpaths[new].append(w)
                                duped = True
                                break
                if not duped:
                    if k not in subpaths:
                        subpaths[k] = []
                    subpaths[k].append(v)

            if len(subpaths) > 0:
                for k,sp in subpaths.items():
                    if k in new_tree:
                        new_tree[k] = new_tree[k] + sp
                    else:
                       new_tree[k] = sp

        return {k:tuple(v) for k,v in new_tree.items()}


    @classmethod
    def generate_nquanta_filter(cls, initials, rules, finals):
        """
        Takes the initial number of quanta, a set of possible rules, and
        a set of final numbers of quanta and determines which rules apply

        :param initial:
        :type initial:
        :param rules:
        :type rules:
        :return:
        :rtype:
        """

        return {
            k: [r for r in rules if k + sum(r)] in finals for k in initials
        }



import itertools

import numpy as np, enum, abc, scipy.sparse as sp
from McUtils.Combinatorics import SymmetricGroupGenerator, PermutationRelationGraph
import McUtils.Numputils as nput
from ..BasisReps import BasisStateSpace, BasisMultiStateSpace, SelectionRuleStateSpace, BraKetSpace, HarmonicOscillatorProductBasis

__all__ = [
    "DegeneracySpec",
    "DegenerateMultiStateSpace"
]

__reload_hook__ = ["..BasisReps"]

class DegenerateSpaceInputFormat(enum.Enum):
    Groups = "groups"
    QuantaSpecRules = "nT"
    Polyads = "polyads"
    Pairs = "pairs"
    StrongCouplings = "wfc_threshold"
    EnergyCutoff = "energy_window"
    Callable = "callable"
    MartinTest = "martin_threshold"

class DegeneracySpec(metaclass=abc.ABCMeta):
    """
    Provides a container for specifying degeneracies
    in a way that can be cleanly canonicalized
    """

    application_order = 'pre'
    group_filter = None
    def __init__(self,
                 application_order=None, group_filter=None, energy_cutoff=2.25e-3,
                 test_modes=None,
                 maximize_filtered_groups=True,
                 decoupling_overide=100,
                 extra_groups=None,
                 inconsistent_polyads=None
                 ):
        if application_order is None:
            application_order = self.application_order
        self.application_order = application_order
        self.energy_cutoff = energy_cutoff
        if group_filter is None:
            group_filter = self.group_filter
        self.group_filter = group_filter
        self.test_modes = test_modes
        self.maximize_filtered_groups = maximize_filtered_groups
        self.decoupling_overide = decoupling_overide
        self.extra_groups = extra_groups
        self.inconsistent_polyads = inconsistent_polyads

    repr_opts = ['energy_cutoff']
    def __repr__(self):
        return "{}({})".format(type(self).__name__, ", ".join('{}={}'.format(k, getattr(self, k)) for k in self.repr_opts))
    @classmethod
    def merge_state_blocks(cls, state_blocks):
        return [g[1] for g in PermutationRelationGraph.make_relation_graph(state_blocks)[0]]
    def get_degenerate_group_filter(self, solver, corrs=None, threshold=None):
        group_filter = self.group_filter
        if isinstance(group_filter, str) and group_filter == 'default':
            group_filter = None
        if group_filter is None:
            zero_order_energy_cutoff = self.energy_cutoff
            test_modes = self.test_modes
            if test_modes is None:
                test_modes = solver.high_frequency_modes
            if len(test_modes) == 0:
                test_modes = None
            group_filter = dict(
                corrections=corrs,
                energy_cutoff=zero_order_energy_cutoff,
                energies=solver.zero_order_energies,
                threshold=threshold,
                target_modes=test_modes,
                maximize_groups=self.maximize_filtered_groups,
                decoupling_overide=self.decoupling_overide,
                extra_groups=self.extra_groups
            )
        elif isinstance(group_filter, str) and group_filter == 'unfiltered':
            group_filter = None
        return group_filter

    format = None
    @classmethod
    def get_format_mapping(cls):
        return {
            DegenerateSpaceInputFormat.Groups: GroupsDegeneracySpec,
            DegenerateSpaceInputFormat.QuantaSpecRules: TotalQuantaDegeneracySpec,
            DegenerateSpaceInputFormat.Polyads: PolyadDegeneracySpec,
            DegenerateSpaceInputFormat.StrongCouplings: StronglyCoupledDegeneracySpec,
            DegenerateSpaceInputFormat.EnergyCutoff: EnergyCutoffDegeneracySpec,
            DegenerateSpaceInputFormat.Callable: CallableDegeneracySpec,
            DegenerateSpaceInputFormat.MartinTest: MartinTestDegeneracySpec,
        }

    @classmethod
    def get_default_spec(cls):
        return StronglyCoupledDegeneracySpec()
    @classmethod
    def from_spec(cls, spec, format=None, **kwargs):
        if spec is False:
            spec = None
        if spec is None:
            return None
        elif (
                isinstance(spec, str) and spec == 'auto'
                or spec is True
        ):
            return cls.get_default_spec()

        if format is None:
            (spec, opts), format = cls.infer_format(spec)
            if opts is not None:
                kwargs = dict(opts, **kwargs)
        res_cls = cls.get_format_mapping()[format]

        return res_cls(spec, **kwargs)

    @staticmethod
    def _key(obj, key):

        try:
            val = getattr(obj, key)
        except AttributeError:
            val = None
        if val is None:
            try:
                val = obj[key]
            except (TypeError, KeyError, AttributeError, IndexError):
                val = None
        return val


    @classmethod
    def infer_format(cls, spec):

        if spec is None:
            return (None, None), None

        fmt = cls._key(spec, 'format')
        if fmt is not None:
            return spec, fmt

        for fmt in DegenerateSpaceInputFormat:
            key = fmt.value
        #     ['states', DegenerateSpaceInputFormat.Groups],
        #     ['nT', DegenerateSpaceInputFormat.QuantaSpecRules],
        #     ['energy_cutoff', DegenerateSpaceInputFormat.EnergyCutoff],
        #     ['polyads', DegenerateSpaceInputFormat.Polyads],
        #     ['couplings', DegenerateSpaceInputFormat.StrongCouplings],
        #     ['callable', DegenerateSpaceInputFormat.Callable]
        # ]:
            val = cls._key(spec, key)
            if val is not None:
                opts = cls._key(spec, 'options')
                if opts is None and isinstance(spec, dict):
                    opts = spec.copy()
                    del opts[key]
                return (val, opts), fmt
        else:
            for fmt,kls in cls.get_format_mapping().items():
                new_spec = kls.canonicalize(spec)
                if new_spec is not None:
                    return (new_spec, None), fmt
            else:
                return (None, None), None

    @abc.abstractmethod
    def get_groups(self, input_states, solver=None, **kwargs):
        """
        :param solver:
        :type solver:
        :param input_states:
        :type input_states:
        :return:
        :rtype:
        """

        raise NotImplementedError("abstract interface")

    @staticmethod
    def get_group_polyad_relation(exc):
        rows, cols = np.triu_indices(len(exc), k=1)
        diffs = exc[cols] - exc[rows]
        negs = diffs < 0
        a = diffs
        b = np.zeros_like(a)
        b[negs] = -diffs[negs]
        a[negs] = 0
        return np.moveaxis(np.array([a, b]), 0, 1)

    @classmethod
    def get_polyad_pairs_from_polyad_specs(cls, polyads):
        return [
            [p1, p2]
            for polyad in polyads
            for i,p1 in enumerate(polyad)
            for p2 in polyad[i+1:]
        ]
    def get_polyad_pairs(self, input_states, solver=None, **kwargs):
        """
        :param solver:
        :type solver:
        :param input_states:
        :type input_states:
        :return:
        :rtype:
        """
        groups = self.get_groups(input_states, solver=solver, **kwargs)
        if len(groups) == 0:
            base_polyads = []
        else:
            base_polyads = np.unique(
                np.concatenate(
                    [
                        self.get_group_polyad_relation(g)
                        for g in groups
                    ],
                    axis=0
                ),
                axis=0
            )

        # filter out anything that just adds one quantum
        base_polyads = [
            b for b in base_polyads
            if (np.sum(np.abs(b[0])) > 0 and np.sum(np.abs(b[1])) > 0)
        ]

        return self.get_polyad_pairs_from_polyad_specs(base_polyads)

    @classmethod
    @abc.abstractmethod
    def canonicalize(cls, spec):
        raise NotImplementedError("abstract interface")

    def get_group_filter(self, **kwargs):
        return DegenerateMultiStateSpace.construct_filer(**kwargs)

class EnergyCutoffDegeneracySpec(DegeneracySpec):
    """

    """

    format = DegenerateSpaceInputFormat.EnergyCutoff
    def __init__(self, cutoff, **opts):
        super().__init__(**opts)
        self.cutoff = cutoff

    @classmethod
    def canonicalize(cls, spec):
        return isinstance(spec, (int, float, np.integer, np.floating))

    @classmethod
    def _group_states_by_energy_cutoff(cls, H0, states, cutoff):
        """
        :type H: Iterable[SparseArray]
        :type states: BasisStateSpace
        :type total_state_space: BasisMultiStateSpace
        :type cutoff: float
        :rtype: Iterable[BasisStateSpace]
        """
        # we look for states with energies within a range...
        # so initially we pull the sets of energies

        diag_inds = BraKetSpace(states, states)
        energies = H0[diag_inds, diag_inds]
        degenerate_groups = []
        # then we look through the input states
        for n, e in enumerate(energies):
            # we only want to apply this once per degenerate group
            # NOTE: this is a path to subtlety, since
            #   if state a is within 50 cm^-1 of state b, and state b is within of c,
            #   you might argue a and c are degenerate
            #   we are wagering that states are distinct _enough_ such that this is not
            #   an issue, but if it is a different strategy will be required
            if all(n not in d for d in degenerate_groups):
                e_diffs = np.abs(energies - e)
                inds = np.where(e_diffs < cutoff)[0]
                degenerate_groups.append(set(inds))
        # raise Exception(degenerate_groups)
        degenerate_groups = [states.take_subspace(np.array(list(d), dtype=int)) for d in degenerate_groups]
        return degenerate_groups

class MartinTestDegeneracySpec(DegeneracySpec):
    application_order = 'mid'
    group_filter = 'default'
    def __init__(self, threshold=4.6e-6, test_energy_window=4.6e-3, convert=True, frequencies=None, **opts):
        super().__init__(**opts)
        if convert and threshold > 1e-2: # help coerce it into Hartrees
            threshold = threshold / 219475 # only need a rough conversion...
        self.threshold = threshold
        self.window = test_energy_window
        self.frequencies = frequencies
        self._test_groups = None
        self._states = None
        self._basis = None
        self._matrix = None

    repr_opts = ['energy_cutoff', 'threshold']
    def prep_states(self, states:BasisStateSpace):
        state_list = states.excitations
        zo_engs = np.dot(state_list, self.frequencies)
        extended_states = states.basis.operator('x', 'x', 'x').get_transformed_space(states)
        key_inds = []
        exc_groups = []
        for i, (s, e0, exc) in enumerate(zip(state_list, zo_engs, extended_states)):
            e1 = np.dot(exc.excitations, self.frequencies)
            eng_diffs = e1 - e0
            pos = np.where(np.abs(eng_diffs < self.window))
            if len(pos) > 0 and len(pos[0]) > 0:
                key_inds.append(i)
                exc_groups.append(exc.take_subspace(pos[0]))
        if len(key_inds) > 0:
            self._test_groups = SelectionRuleStateSpace(
                states.take_subspace(key_inds),
                exc_groups
            )
            new = states.union(self._test_groups.to_single().take_unique())
        else:
            new = states
            self._test_groups = []
        return new
    def get_coupled_spaces(self, input_states:BasisStateSpace, solver=None):
        # raise NotImplementedError("...")
        if self._test_groups is None:
            self._test_groups = self.prep_states(input_states)
        elif len(self._test_groups) == 0:
            return []
        if solver is None:
            raise ValueError("need a solver with fully evaluated perturbations first...")
        total_basis = solver.flat_total_space
        # engs = solver.zero_order_energies
        h1 = solver.representations[1]
        groups = self._test_groups #type:SelectionRuleStateSpace
        keys = groups.representative_space
        key_inds = total_basis.find(keys)
        spaces = []
        martin_inds = [[], []]
        martin_vals = []
        for i, (k, subspace) in enumerate(zip(key_inds, groups.flat)):
            e0 = np.dot(keys.excitations[i], self.frequencies)
            e1 = np.dot(subspace.excitations, self.frequencies)
            p = total_basis.find(subspace)
            e_diffs = np.abs(e1 - e0)
            repr_elems = h1[k, p].asarray()
            # diffs = subspace.excitations - keys.excitations[np.newaxis, k]
            # diff_sums = np.sum(np.abs(diffs), axis=1)
            # raise Exception(diffs, keys.excitations[np.newaxis, k], subspace.excitations)
            # weights = 1
            test_val = (repr_elems**4)/(e_diffs**3)
            # if (keys.excitations[i] == np.array([0, 0, 0, 1, 0, 0, 0, 0, 0])).all():
            #     raise Exception(test_val*219465, repr_elems, e_diffs, subspace.excitations)
            # if (keys.excitations[i] == np.array([0, 0, 0, 0, 0, 0, 1, 0, 0])).all():
            #     raise Exception(test_val*219465, repr_elems, e_diffs, subspace.excitations)
            martin_inds[0].extend([i]*len(subspace))
            martin_inds[1].extend(p)
            martin_vals.extend(test_val)
            pos = np.where(test_val > self.threshold)
            if len(pos) > 0 and len(pos[0]) > 0:
                # print(pos[0], subspace.excitations)
                new = np.concatenate([
                        keys.excitations[i:i+1],
                        subspace.excitations[pos[0],]#.take_subspace(pos[0]).excitations
                    ], axis=0)
                # print(keys.excitations[i:i+1], new)
                new_sums = np.sum(new, axis=1)
                new = new[new_sums > 0]
                spaces.append(new)

        self._states = keys
        self._basis = solver.flat_total_space
        self._matrix = nput.SparseArray.from_data(
            (
                np.array(martin_vals),
                tuple(np.array(x) for x in martin_inds)
            ),
            shape=(len(self._states), len(self._basis))
        )
        # raise Exception("...")
        return spaces
    # @classmethod
    def get_groups(self, input_states:BasisStateSpace, solver=None, extra_groups=None, **kwargs):
        groups = self.get_coupled_spaces(input_states, solver=solver)
        indexer = SymmetricGroupGenerator(input_states.ndim)
        if extra_groups is None:
            extra_groups = self.extra_groups
        if extra_groups is None:
            extra_groups = []
        extra_groups = [(indexer.to_indices(g), np.asanyarray(g)) for g in extra_groups]
        groups = PermutationRelationGraph.merge_groups([(indexer.to_indices(g), g) for g in groups] + extra_groups)
        return [g[1] for g in groups]

    @classmethod
    def canonicalize(cls, spec):
        return isinstance(spec.threshold, float)

    def get_group_filter(self, threshold=None, states=None, basis=None, corrections=None, **kwargs):
        return DegenerateMultiStateSpace.construct_filer(
            states=self._states,
            basis=self._basis,
            corrections=self._matrix,
            threshold=self.threshold,
            threshold_step_size=1/219475,
            extra_groups=self.extra_groups,
            **kwargs
        )

class StronglyCoupledDegeneracySpec(DegeneracySpec):
    """

    """
    # application_order = 'post'
    format = DegenerateSpaceInputFormat.StrongCouplings
    default_threshold=.3
    def __init__(self, wfc_threshold=None, state_filter=None, extend_spaces=True, iterations=None,
                 evaluator=None,
                 **opts):
        self.evaluator = evaluator
        self._couplings = {}
        super().__init__(**opts)
        if wfc_threshold is None or isinstance(wfc_threshold, str) and wfc_threshold == 'auto':
            wfc_threshold = self.default_threshold
        self.wfc_threshold = wfc_threshold
        self.state_filter = state_filter
        self.extend_spaces=extend_spaces
        self._iterations = iterations
        self.iterations = iterations

    @property
    def application_order(self):
        return 'post' if self.evaluator is None else 'pre'
    @application_order.setter
    def application_order(self, ord):
        ...

    repr_opts = ['energy_cutoff', 'wfc_threshold']
    def prep_states(self, input_states):
        return input_states

    def identify_strong_couplings(self, solver, corrs):
        degenerate_correction_threshold = self.wfc_threshold
        test_modes = self.test_modes
        if test_modes is None:
            test_modes = solver.high_frequency_modes

            # raise Exception(test_modes, test_freqs * 219465, fundamentals.excitations, where)

        state_filter = self.state_filter
        zero_order_energy_cutoff = self.energy_cutoff
        if state_filter is None:
            state_filter = lambda state, couplings: corrs.default_state_filter(state, couplings,
                                                                               energy_cutoff=zero_order_energy_cutoff,
                                                                               energies=solver.zero_order_energies,
                                                                               basis=solver.flat_total_space,
                                                                               target_modes=test_modes
                                                                               )
        sc = corrs.find_strong_couplings(threshold=degenerate_correction_threshold, state_filter=state_filter)
        return sc

    def get_input_state_couplings(self, input_states):
        inds = input_states.indices
        needs_couplings = [
            n for n,i in enumerate(inds)
            if i not in self._couplings
        ]
        if len(needs_couplings) > 0:
            needs_coupling_states = input_states.take_subspace(needs_couplings)
            wfcs = self.evaluator.get_test_wfn_corrs(needs_coupling_states, self.energy_cutoff)
            if wfcs is not None:
                for input_state, final_state, corrs in zip(wfcs.initial_states.indices, wfcs.final_states, wfcs.corrections):
                    coupling_pos = np.where(np.abs(corrs) > self.wfc_threshold)
                    if len(coupling_pos) > 0 and len(coupling_pos[0]) > 0:
                        coupling_pos = coupling_pos[1]
                        self._couplings[input_state] = final_state.take_subspace(coupling_pos)
            for i in inds:
                if i not in self._couplings:
                    self._couplings[i] = None
        return {
            i:self._couplings[i]
            for i in inds
            if self._couplings[i] is not None
        }

    def get_groups(self, input_states, couplings=None, solver=None, extra_groups=None, **kwargs):
        """
        :param input_states:
        :type input_states:
        :param solver:
        :type solver:
        :return:
        :rtype:
        """
        if self.evaluator is not None:
            couplings = self.get_input_state_couplings(input_states)
            # if len(couplings) == 0:
            #     return None
        else:
            if couplings is None:
                raise ValueError("need couplings")
        if extra_groups is None:
            extra_groups = self.extra_groups
        return self.get_strong_coupling_space(input_states, couplings, extra_groups=extra_groups)

    @classmethod
    def canonicalize(cls, spec):
        if isinstance(spec, (int, float, np.integer, np.floating)):
            return spec
        else:
            try:
                return {k:GroupsDegeneracySpec._validate_grp(v) for k,v in spec.items()}
            except (np.VisibleDeprecationWarning, AttributeError, StopIteration):
                return None

    @classmethod
    def get_strong_coupling_space(cls, states: BasisStateSpace, couplings: dict, extra_groups=None):
        indexer = SymmetricGroupGenerator(states.ndim)
        groups = [None] * len(states.indices)
        for n, i in enumerate(states.indices):
            if i in couplings:
                groups[n] = [i] + list(couplings[i].indices)
            else:
                groups[n] = [i]

        if extra_groups is None:
            extra_groups = []
        extra_groups = [(indexer.to_indices(g), np.asanyarray(g)) for g in extra_groups]

        # print("premerge groups:", groups)

        groups = PermutationRelationGraph.merge_groups(
            [(g, indexer.from_indices(g)) for g in groups] + extra_groups
        )

        # print("????", [g[0] for g in groups]) # need to track down why we're getting a negative index here
                              # presumably from an integer overflow in a int8 or a conversion
                              # of some sort...

        return [g[1] for g in groups]

class GroupsDegeneracySpec(DegeneracySpec):
    """

    """

    format = DegenerateSpaceInputFormat.Groups
    def __init__(self, groups, **opts):
        super().__init__(**opts)
        self.groups = groups

    @staticmethod
    def _validate_grp(grp):
        try:
            b = grp[0]
        except (IndexError, TypeError):
            raise StopIteration("malformatted")
        else:
            return np.asanyarray(grp)
    @classmethod
    def canonicalize(cls, spec):
        try:
            return [cls._validate_grp(s) for s in spec]
        except (np.VisibleDeprecationWarning, AttributeError, StopIteration):
            return None

    def get_groups(self, input_states, solver=None, **kwargs):
        """
        :param solver:
        :type solver:
        :param input_states:
        :type input_states:
        :return:
        :rtype:
        """
        # do I want to canonicalize if group size is too small?
        return self.groups

class PolyadDegeneracySpec(DegeneracySpec):
    """

    """
    format = DegenerateSpaceInputFormat.Polyads
    def __init__(self, polyads,
                 max_quanta=None,
                 iterations=2,
                 require_converged=False,
                 extra_groups=None,
                 extra_polyads=None,
                 full_group_polyads=True,
                 **opts
                 ):
        super().__init__(**opts)
        self.polyads = polyads
        self.max_quanta = max_quanta
        self.iterations = iterations
        self.require_converged = require_converged
        self.extra_groups = extra_groups
        self.full_group_polyads = full_group_polyads

    @staticmethod
    def _validate_rule(pair):
        try:
            if len(pair) != 2:
                raise StopIteration("malformatted")
            b = pair[0]
        except (IndexError, TypeError):
            raise StopIteration("malformatted")
        else:
            if len(pair[0]) == len(pair[1]):
                return pair
            else:
                raise StopIteration("malformatted")
    @classmethod
    def _check_rule(cls, pair):
        try:
            cls._validate_rule(pair)
        except StopIteration:
            return False
        else:
            return True
    @classmethod
    def canonicalize(cls, spec):
        try:
            return [cls._validate_rule(s) for s in spec]
        except (np.VisibleDeprecationWarning, AttributeError, StopIteration):
            return None

    def get_groups(self, input_states, solver=None, **kwargs):
        """
        :param solver:
        :type solver:
        :param input_states:
        :type input_states:
        :return:
        :rtype:
        """
        if isinstance(input_states, BasisStateSpace):
            input_states = input_states.excitations
        return self.get_degenerate_polyad_space(
            input_states,
            self.polyads,
            max_quanta=self.max_quanta,
            iterations=self.iterations,
            require_converged=self.require_converged,
            extra_groups=self.extra_groups
        )
    @classmethod
    def get_degenerate_polyad_space(cls, states, polyadic_pairs,
                                    max_quanta=None, iterations=2,
                                    require_converged=False, extra_groups=None
                                    ):
        """
        Gets degenerate spaces by using pairs of transformation rules to
        take an input state and connect it to other degenerate states

        :param states: the input states
        :type states:
        :param polyadic_pairs: the transformation rules
        :type polyadic_pairs:
        :param max_quanta: the max quanta to allow in connected states
        :type max_quanta:
        :return:
        :rtype:
        """

        if cls._check_rule(polyadic_pairs[0]): # we only have one space
            polyadic_pairs = [polyadic_pairs]
        if isinstance(iterations, int):
            iterations = [iterations]*len(polyadic_pairs)

        groups = None
        for polyad, iterations in zip(polyadic_pairs, iterations):
            graph = PermutationRelationGraph(polyad)
            if groups is None:
                groups = graph.build_state_graph(states,
                                                 extra_groups=extra_groups,
                                                 max_sum=max_quanta,
                                                 max_iterations=iterations,
                                                 raise_iteration_error=require_converged
                                                 )
            else:
                _ = []
                for states in groups:
                    sub = np.concatenate(graph.build_state_graph(states,
                                                     extra_groups=extra_groups,
                                                     max_sum=max_quanta,
                                                     max_iterations=iterations,
                                                     raise_iteration_error=require_converged
                                                     ))
                    _.append(sub)
                groups = _
        groups = [g for g in groups if len(g) > 1]
        return groups



    def get_polyad_pairs(self, input_states, solver=None, full_group_polyads=None, **kwargs):
        if full_group_polyads is None:
            full_group_polyads = self.full_group_polyads
        if full_group_polyads:
            return super().get_polyad_pairs(input_states, solver=solver, **kwargs)
        else:
            return self.get_polyad_pairs_from_polyad_specs(self.polyads)

    @staticmethod
    def _is_polyad_rule(d, n_modes):
        try:
            return (
                    len(d) == 2
                    and len(d[0]) == n_modes
                    and len(d[1]) == n_modes
            )
        except TypeError:
            return False

class TotalQuantaDegeneracySpec(PolyadDegeneracySpec):
    """

    """

    format = DegenerateSpaceInputFormat.QuantaSpecRules
    def __init__(self, n_T_vectors, max_quanta=3, target_modes=None, extra_polyads=None, **opts):
        self.max_quanta = max_quanta
        self.target_modes = target_modes
        self.nt_vecs = [n_T_vectors] if isinstance(n_T_vectors[0], int) else n_T_vectors
        if extra_polyads is None:
            extra_polyads = []
        if len(extra_polyads) > 0 and self._check_rule(extra_polyads[0]):
            extra_polyads = [extra_polyads]
        rules = [
                    self.make_nt_polyad(v, max_quanta=max_quanta) for v in self.nt_vecs
                ] + extra_polyads
        # raise Exception(rules)
        super().__init__(rules, **opts)

    @classmethod
    def reduce_rule(cls, a, b):
        a = a.copy()
        b = b.copy()

        m1 = np.min(a[a>0])
        m2 = np.min(b[b>0])
        if m2 > m1:
            m1, m2 = m2, m1
        if m1 % m2 == 0:
            a = a//m2
            b = b//m2
        reduce_pos = np.where(np.logical_and(a > 0, b > 0))
        if len(reduce_pos) > 0:
            reduce_pos = reduce_pos[0]
            if len(reduce_pos) > 0:
                pairs = np.concatenate([a[reduce_pos][:, np.newaxis], b[reduce_pos][:, np.newaxis]], axis=1)
                min_vals = np.min(pairs, axis=1)
                a[reduce_pos] = a[reduce_pos] - min_vals
                b[reduce_pos] = b[reduce_pos] - min_vals
        return a,b

    @classmethod
    def make_nt_polyad(cls, nt, target_modes=None, max_quanta=3):  # this should be of the form num of quanta give same E
        # we want to choose different subsets of the nT vector to construct rules
        # we'll do this in a kinda dumb way where we force a max number of quanta for these polyad
        # rules and explicitly test states of a given form
        ndim = len(nt)
        nt = np.asanyarray(nt)
        if target_modes is None:
            # target_modes = np.arange(len(nt))
            target_modes = np.where(nt != 0)[0]

        nt = nt[target_modes]
        states = BasisStateSpace.from_quanta(
            HarmonicOscillatorProductBasis(len(nt)),
            list(range(1, max_quanta+1))
        ).excitations

        nts = np.dot(states, nt)
        (keys, groups), _ = nput.group_by(states, nts)
        rules = set()
        for k,g in zip(keys, groups):
            if k > 0:
                for n,i in enumerate(g):
                    for j in g[n+1:]:
                        a, b = cls.reduce_rule(i, j)
                        # insert zeros at dropped positions
                        if len(target_modes) < ndim:
                            _ = np.zeros(ndim, dtype=int)
                            _[target_modes] = a
                            a = _
                            _ = np.zeros(ndim, dtype=int)
                            _[target_modes] = b
                            b = _
                        a = tuple(a)
                        b = tuple(b)
                        rules.add((a, b))
        return list(rules)

class CallableDegeneracySpec(DegeneracySpec):
    """

    """

    format = DegenerateSpaceInputFormat.Callable
    def __init__(self, callable, **opts):
        super().__init__(**opts)
        self.callable = callable

    def get_groups(self, input_states, solver=None, **kwargs):
        """
        :param input_states:
        :type input_states:
        :param solver:
        :type solver:
        :return:
        :rtype:
        """
        return self.callable(input_states, solver=solver, **kwargs)

class DegenerateMultiStateSpace(BasisMultiStateSpace):

    @staticmethod
    def default_group_filter(group,
                             states=None,
                             basis=None,
                             corrections=None,
                             threshold=None,
                             energy_cutoff=None,
                             energies=None,
                             decoupling_overide=10,
                             maximize_groups=True,
                             target_modes=None,
                             threshold_step_size=.1,
                             extra_groups=None
                             ):
        """
        Excludes modes that differ in only one position, prioritizing states with fewer numbers of quanta
        (potentially add restrictions to high frequency modes...?)

        :param input_state:
        :type input_state:
        :param couplings:
        :type couplings:
        :return:
        :rtype:
        """

        if len(group) < 2:
            return group

        if target_modes is not None:
            sub = group.take_subdimensions(target_modes)
            inds = sub.indices
            exc = sub.excitations
        else:
            inds = group.indices
            exc = group.excitations

        if corrections is not None:
            if isinstance(corrections, np.ndarray):
                state_inds = basis.find(group)
                group_inds = state_inds
                found_pos = state_inds < len(states)
            elif hasattr(corrections, 'total_basis'): # PT corrections
                states = corrections.states
                basis = corrections.total_basis
                group_inds = basis.find(group)
                state_inds = states.find(group, dtype=int, missing_val=len(states))
                found_pos = state_inds < len(states)
                kill_spots = (np.arange(len(group_inds))[found_pos], state_inds[found_pos])
                corr_mat = [np.abs(x[:, group_inds].asarray().T) for x in corrections.wfn_corrections[1:]]
                max_corr = None
                for c in corr_mat:
                    c[kill_spots] = 0
                    if max_corr is None:
                        max_corr = np.max(c, axis=1)
                    else:
                        submax = np.max(c, axis=1)
                        p = submax > max_corr
                        max_corr[p] = submax[p]
                corrections = sum(corr_mat)
                if threshold is not None:
                    sorting = np.arange(len(max_corr))
                else:
                    sorting = np.argsort(-max_corr - 1 * (found_pos))  # prioritize states inside requested packet for marginal cases
            elif basis is not None and states is not None:
                group_inds = basis.find(group)
                state_inds = states.find(group, dtype=int, missing_val=len(states))
                found_pos = state_inds < len(states)
                found_states = state_inds[found_pos]
                kill_spots = (np.arange(len(group_inds))[found_pos], found_states)
                corrections = np.abs(corrections[:, group_inds].asarray().T)
                max_corr = None
                corrections[kill_spots] = 0
                if max_corr is None:
                    max_corr = np.max(corrections, axis=1)
                else:
                    submax = np.max(corrections, axis=1)
                    p = submax > max_corr
                    max_corr[p] = submax[p]
                if threshold is not None:
                    sorting = np.arange(len(max_corr))
                else:
                    sorting = np.argsort(-max_corr - 1*found_pos)  # prioritize states inside requested packet for marginal cases

        else:
            sorting = np.argsort(inds)
            corrections = None

        exc = exc[sorting]
        diffs = exc[:, np.newaxis] - exc[np.newaxis, :]  # difference matrix (s, s, m)
        tots = exc[:, np.newaxis] + exc[np.newaxis, :]  # difference matrix (s, s, m)
        diff_sums = np.sum(np.abs(diffs), axis=2)  # (s, s)
        tots_sums = np.sum(np.abs(tots), axis=2)  # (s, s)
        elims = []
        if threshold is not None and corrections is not None:
            if energy_cutoff is None:
                e_vec = None
                bad_pos = np.where(np.logical_and(
                    diff_sums == 1,
                    tots_sums > 0
                ))  # np.logical_or(diff_sums == 1, diff_sums == 0))
            else:
                e_vec = energies[group_inds]
                bad_pos = np.where(np.abs(np.subtract.outer(e_vec, e_vec)) > energy_cutoff)
            if len(bad_pos) > 0 and len(bad_pos[0]) > 0:
                N = len(group_inds)
                base_corr = corrections
                base_vals = base_corr[:, state_inds[found_pos]]
                base_mat = np.zeros((N, N), dtype=float)
                base_mat[:, np.where(found_pos)[0]] = base_vals
                base_mat = (base_mat.T + base_mat) - 2*np.diag(np.diag(base_mat))
                # if decoupling_overide > 0:
                #     truly_bad = base_mat[bad_pos] < decoupling_overide
                # else:
                #     truly_bad = np.full(len(base_mat[bad_pos]), True)
                # # introduce an override so things that are super strongly coupled can remain coupled
                bad_pos = tuple(b for b in bad_pos)
                base_mat[bad_pos] = 0  # refuse to couple problem states
                base_mat[bad_pos[1], bad_pos[0]] = 0  # refuse to couple problem states

                def get_groups(cm):
                    _, labs = sp.csgraph.connected_components(cm, return_labels=True)
                    (grp_keys, groups), _ = nput.group_by(np.arange(N), labs)
                    A = np.zeros((N, N))
                    for g in groups:
                        A[np.ix_(g, g)] = 1
                    return groups, A

                corr_mat = (base_mat > threshold).astype(int)
                groups, A = get_groups(corr_mat)

                target_threshold = threshold
                step_size = threshold_step_size
                while (A[bad_pos] > 0).any() and decoupling_overide > threshold:
                    threshold = threshold + step_size
                    corr_mat = (base_mat > threshold).astype(int)
                    groups, A = get_groups(corr_mat)

                if maximize_groups or (decoupling_overide > 0 and decoupling_overide < threshold):
                    base_thresh = target_threshold
                    if not maximize_groups:
                        base_thresh = decoupling_overide
                    dropped_pos = np.where(np.logical_and(base_mat > base_thresh, base_mat < threshold))
                    if len(dropped_pos) > 0:
                        # figure out if any can be added without causing breakdowns
                        sorting = np.argsort(-base_mat[dropped_pos])
                        dropped_pos = tuple(d[sorting] for d in dropped_pos)
                        for i,j in zip(*dropped_pos): # this could be very expensive...
                            if e_vec is not None:
                                # we can maybe take a shortcut on figuring out if we can extend the group
                                g_i = None
                                g_j = None
                                # first we check if we have a group with `i` or `j` in it
                                for n,g in enumerate(groups):
                                    ix,jx = np.searchsorted(g, [i, j])
                                    if ix < len(g) and g[ix] == i:
                                        g_i = [n, g]
                                    if jx < len(g) and g[jx] == j:
                                        g_j = [n, g]
                                    if g_i is not None and g_j is not None: # need proper graph tools
                                        break
                                else:
                                    # we're just adding to a group, not merging
                                    if g_j is not None:
                                        g_i, g_j = g_j, g_i
                                        i, j = j, i
                                    if g_i is not None:
                                        e_g = e_vec[g_i[1]]
                                        w = max(abs(e_vec[i] - np.min(e_g)), abs(e_vec[i] - np.max(e_g)))
                                        if w < energy_cutoff:
                                            groups[g_i[0]] = np.sort(np.concatenate([g_i[1], [j]]))
                                        continue
                                    # else: # should never matter: would mean states were dropped but are within the energy window

                            corr_mat[i, j] = base_mat[i, j]
                            _, A = get_groups(corr_mat)
                            if (A[bad_pos] > 0).any():
                                corr_mat[i, j] = 0.
                            else:
                                groups = _

                    if (
                            (decoupling_overide > 0 and decoupling_overide < threshold)
                            or extra_groups is not None
                    ):
                        # states with blowups that we need to reinsert
                        if (decoupling_overide > 0 and decoupling_overide < threshold) is not None:
                            need_pos = np.where(np.logical_and(base_mat > decoupling_overide, base_mat < threshold))
                        else:
                            need_pos = ()
                        if extra_groups is not None:
                            for g in extra_groups:
                                try:
                                    new_inds = group.find(g)
                                except IndexError:
                                    pass
                                else:
                                    new_inds = list(itertools.combinations(new_inds, 2))
                                    new_inds = ([x[0] for x in new_inds], [x[1] for x in new_inds])
                                    if len(need_pos) > 0:
                                        need_pos = (
                                            np.concatenate([need_pos[0], new_inds[0]]),
                                            np.concatenate([need_pos[1], new_inds[1]]),
                                        )
                                    else:
                                        need_pos = new_inds
                        if len(need_pos) > 0:
                            sorting = np.argsort(-base_mat[need_pos])
                            need_pos = tuple(d[sorting] for d in need_pos)
                            for i,j in zip(*need_pos):
                                # we try to find a group to insert into
                                g_i = None
                                g_j = None
                                # first we check if we have a group with `i` or `j` in it
                                for n,g in enumerate(groups):
                                    ix,jx = np.searchsorted(g, [i, j])
                                    if ix < len(g) and g[ix] == i:
                                        g_i = [n, g]
                                    if jx < len(g) and g[jx] == j:
                                        g_j = [n, g]
                                    if g_i is not None and g_j is not None:
                                        if g_i[0] == g_j[0]: # nothing needed to do
                                            break
                                        # we move j from its group into i's group
                                        # or vice versa
                                        i_couplings = base_mat[i, g_i[1]]
                                        j_couplings = base_mat[j, g_j[1]]
                                        if np.linalg.norm(j_couplings) > np.linalg.norm(i_couplings):
                                            # swap which gets moved
                                            g_i, g_j = g_j, g_i
                                            i, j = j, i
                                        groups[g_i[0]] = np.sort(np.concatenate([g_i[1], [j]]))
                                        groups[g_j[0]] = np.setdiff1d(g_j[1], [j])
                                        break
                                else:
                                    # we're just adding to a group, not merging
                                    if g_j is not None:
                                        g_i, g_j = g_j, g_i
                                        i, j = j, i
                                    if g_i is not None:
                                        groups[g_i[0]] = np.sort(np.concatenate([g_i[1], [j]]))
                                        continue
                                    else: # I don't think we'll ever get here but...?
                                        groups.append(np.array([i, j]))

                group = [group.take_subspace(g) for g in groups if len(g) > 1]

        else:
            for n in np.arange(len(exc)):
                diff_vec = diff_sums[n]
                bad_pos = np.where(diff_vec == 1)
                if len(bad_pos) > 0:
                    bad_pos = bad_pos[0]
                    if len(bad_pos) > 0:
                        kills = bad_pos[bad_pos > n]
                        elims.extend(sorting[kills])
                        diff_sums[bad_pos, :] = 0

            if len(elims) > 0:
                mask = np.setdiff1d(np.arange(len(exc)), elims)
                group = group.take_subspace(mask)

        return group

    @classmethod
    def construct_filer(cls, spec=None, **kwargs):
        if spec is not None:
            return spec.get_group_filter(**kwargs)
        else:
            return lambda group:cls.default_group_filter(group, **kwargs)

    @classmethod
    def from_spec(cls,
                  degenerate_states,
                  solver=None,
                  full_basis=None,
                  format=None,
                  group_filter=None,
                  log_groups=False,
                  **kwargs
                  ):
        """
        Generates a DegenerateMultiStateSpace object from a number
        of possible specs

        :param solver: the actual applier of the perturbation theory which makes use of the degenerate states
        :type solver: PerturbationTheorySolver
        :return:
        :rtype:
        """

        logger = solver.logger
        states = solver.states #type: BasisStateSpace

        if degenerate_states is not None:
            if not isinstance(degenerate_states, DegeneracySpec):
                degenerate_states = DegeneracySpec.from_spec(degenerate_states, format=format)
            with logger.block(tag="getting degeneracies"):
                deg_states = degenerate_states.get_groups(states, solver=solver, **kwargs)
                if hasattr(group_filter, 'items'):
                    group_filter = cls.construct_filer(**dict(group_filter, spec=degenerate_states))
                new = GroupsDegeneracySpec.canonicalize(deg_states)
                if new is None:
                    raise ValueError("{} returned invalid degenerate subspace spec {}".format(
                        degenerate_states,
                        deg_states
                    ))
                else:
                    degenerate_states = new
                    del deg_states

                degenerate_states = [
                    BasisStateSpace(states.basis, s)#.as_sorted(track_indices=False, track_excitations=False)
                    if not isinstance(s, BasisStateSpace) else s
                    for s in degenerate_states
                ]

                # dds = [d.indices for d in degenerate_states]
                # neg_ps = [i for i,d in enumerate(dds) if np.any(d < 0)]

                # print("DD:", neg_ps, [degenerate_states[i].excitations for i in neg_ps])

                if log_groups:
                    with logger.block(tag="initial degenerate groups:"):
                        for g in degenerate_states:
                            if len(g) > 1:
                                logger.log_print(str(g.excitations).splitlines())
                else:
                    logger.log_print(
                        "{n} degenerate state sets found",
                        n=len([x for x in degenerate_states if len(x) > 1]),
                        # s=[x for x in degenerate_states if len(x) > 1]
                    )

        # build groups of degenerate states for use later
        if degenerate_states is None:
            groups = states.split(1) #[[x] for x in states.indices]  # we're gonna loop through this later so why not destructure now...
        else:
            groups = [None] * len(degenerate_states)
            deg_sets = [(set(d.indices),d) for d in degenerate_states]
            for x in states.indices:
                for i, (d,bs) in enumerate(deg_sets):
                    # if i == 11:
                    #     raise Exception(bs.excitations, bs.indices)
                    if x in d:
                        if groups[i] is None:
                            groups[i] = bs.indices
                        # groups[i].append(x)
                        break
                else:
                    groups.append([x])

        # now turn these into proper BasisStateSpace objects so we can work with them more easily
        ugh = []
        for i,g in enumerate(groups):
            if g is None:
                continue #???
            # g = np.sort(np.array(g))
            if not isinstance(g, BasisStateSpace):
                # if np.any(np.asanyarray(g) < 0):
                #     raise Exception(i, g)
                g = BasisStateSpace(states.basis, np.array(g), mode=BasisStateSpace.StateSpaceSpec.Indices, full_basis=full_basis)
            if group_filter is not None:
                g = group_filter(g)
                if isinstance(g, BasisStateSpace):
                    ugh.append(g)
                else:
                    ugh.extend(g)
            else:
                ugh.append(g)

        for n,x in enumerate(states.indices):
            for g in ugh:
                if x in g.indices:
                    break
            else:
                ugh.append(states.take_subspace([n]))

        # now make a numpy array for initialization
        arrs = np.full(len(ugh), None)
        for i, g in enumerate(ugh):
            arrs[i] = g

        # raise Exception(arrs)

        return cls(arrs)



"""
Defines the fundamental interface we expect all molecules to support
without any concrete implementation details so that we can separate property
calculation from explicit molecule instantiation
"""

import abc, typing, numpy as np
from McUtils.Coordinerds import CoordinateSet

__all__ = ["AbstractMolecule"]

class PotentialSurfaceType(typing.Protocol):
    @property
    def surface(self):
        raise NotImplementedError('abstract interface')
    @property
    def derivatives(self):
        raise NotImplementedError('abstract interface')
    def load(self):
        raise NotImplementedError('abstract interface')
    def update(self, val):
        raise NotImplementedError('abstract interface')
    def insert_atoms(self, atoms, coords, where):
        raise NotImplementedError('abstract interface')
    def delete_atoms(self, where):
        raise NotImplementedError('abstract interface')
    def apply_transformation(self, frame):
        raise NotImplementedError('abstract interface')

class DipoleSurfaceType(typing.Protocol):
    @property
    def surface(self):
        raise NotImplementedError('abstract interface')
    @property
    def derivatives(self):
        raise NotImplementedError('abstract interface')
    def load(self):
        raise NotImplementedError('abstract interface')
    def update(self, val):
        raise NotImplementedError('abstract interface')
    def insert_atoms(self, atoms, coords, where):
        raise NotImplementedError('abstract interface')
    def delete_atoms(self, where):
        raise NotImplementedError('abstract interface')
    def apply_transformation(self, frame):
        raise NotImplementedError('abstract interface')

class NormalModesType(typing.Protocol):
    @property
    def modes(self):
        raise NotImplementedError('abstract interface')
    def load(self):
        raise NotImplementedError('abstract interface')
    def update(self, val):
        raise NotImplementedError('abstract interface')
    def insert_atoms(self, atoms, coords, where):
        raise NotImplementedError('abstract interface')
    def delete_atoms(self, where):
        raise NotImplementedError('abstract interface')
    def apply_transformation(self, frame):
        raise NotImplementedError('abstract interface')

class AbstractMolecule(metaclass=abc.ABCMeta):
    """
    A molecule base class to provide type constraints
    """

    @property
    @abc.abstractmethod
    def name(self) -> typing.AnyStr:
        """
        Provides the name of the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def metadata(self) -> typing.Dict:
        """
        Provides the metadta the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def atoms(self) -> typing.Tuple[typing.AnyStr]:
        """
        Provides the atom names for the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def masses(self) -> np.ndarray:
        """
        Provides the masses for the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def atomic_masses(self) -> np.ndarray:
        """
        Provides the atomic masses for the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def coords(self) -> CoordinateSet:
        """
        Provides the atomic coordinates for the molecule
        """
        raise NotImplementedError('abstract class')
    @coords.setter
    def coords(self, new):
        """
        Sets the atomic coordinates for the molecule
        """
        raise NotImplementedError('abstract class')
    @property
    @abc.abstractmethod
    def internal_coordinates(self) -> CoordinateSet:
        """
        Provides the internal coordinates for the molecule
        """
        raise NotImplementedError('abstract class')
    @property
    @abc.abstractmethod
    def center_of_mass(self) -> np.ndarray:
        """
        Provides the center of mass for the molecule
        """
        raise NotImplementedError('abstract class')
    @property
    @abc.abstractmethod
    def inertial_axes(self) -> np.ndarray:
        """
        Provides the principal axes for the molecule
        """
        raise NotImplementedError('abstract class')


    @property
    @abc.abstractmethod
    def bonds(self) -> typing.Iterable[typing.Tuple]:
        """
        Provides the bonds for the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def potential_surface(self) -> PotentialSurfaceType:
        """
        Provides the potential surface for the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def dipole_surface(self) -> DipoleSurfaceType:
        """
        Provides the dipole surface for the molecule
        """
        raise NotImplementedError('abstract class')
    @property
    @abc.abstractmethod
    def normal_modes(self) -> NormalModesType:
        """
        Provides the normal modes for the molecule
        """
        raise NotImplementedError('abstract class')

    @property
    @abc.abstractmethod
    def source_file(self) -> str:
        """
        Provides the source file for the molecule
        """
        raise NotImplementedError('abstract class')

    @abc.abstractmethod
    def copy(self) -> 'typing.Self':
        """
        Copies the molecule
        """
        raise NotImplementedError('abstract class')

    @abc.abstractmethod
    def take_submolecule(self, pos) -> 'typing.Self':
        """
        Takes a submolecule
        """
        raise NotImplementedError('abstract class')



import numpy as np, scipy.linalg as slag, collections

from McUtils.Scaffolding import Logger
from McUtils.Data import UnitsData
from .NormalModes import NormalModes

__all__ = [
    'ObliqueModeGenerator'
]

class ObliqueModeGenerator:

    def __init__(self, f, g, dimensionless=False, sel=None, frequency_scaled=True):
        f = np.asanyarray(f)
        g = np.asanyarray(g)
        if g.ndim == 1: # vector of masses
            g = np.broadcast_to(1/g[:, np.newaxis], (len(g), len(f) // len(g))).flatten()
            g = np.diag(g)
        if sel is not None:
            f = f[sel, :][:, sel]
            g = g[sel, :][:, sel]
        if dimensionless:
            a = np.diag(np.power(np.diag(f) / np.diag(g), 1 / 4))
            ai = np.diag(np.power(np.diag(g) / np.diag(f), 1 / 4))
            f = ai @ f @ ai
            g = a @ g @ a
        self.f = f
        self.g = g
        self.ndim = f.shape[0]

        freqs, inv, modes = NormalModes.get_normal_modes(
            f, g,
            remove_transrot=True,
            dimensionless=frequency_scaled
        )
        inv = inv.T
        modes = modes.T

        rU, a, rM = np.linalg.svd(modes)
        good_a = np.where(a > 1e-8)[0]
        R = (rU[:, good_a] @ rM[good_a, :])

        self.modes = modes
        self.rotation = R
        self.inverse_scaling = modes @ R.T
        self.scaling = R @ inv

    @classmethod
    def from_molecule(cls, mol, dimensionless=True, sel=None, use_internals=None, frequency_scaled=True):
        from ..Molecools import Molecule
        mol = mol # type:Molecule
        if use_internals is None:
            use_internals = mol.internal_coordinates is not None
        if use_internals:
            return cls(
                mol.get_internal_potential_derivatives(2)[1],
                mol.g_matrix,
                dimensionless=dimensionless,
                sel=sel,
                frequency_scaled=frequency_scaled
            )
        else:
            return cls(
                mol.potential_derivatives[1],
                mol.atomic_masses,
                dimensionless=dimensionless,
                sel=sel,
                frequency_scaled=frequency_scaled
            )

    def run(self, scaling_type='normal'):
        u = self.scaling
        ui = self.inverse_scaling
        if scaling_type == 'inverse':
            u, ui = ui, u
        g = u @ self.g @ u.T
        f = ui.T @ self.f @ ui
        return f, g, u, ui




"""
Provides a class for handling a compiled set of atomic data
"""

import os, sys
from McUtils.Data import DataHandler

__all__ = [ "KEData", "KEDataHandler" ]

class KEDataHandler(DataHandler):
    """
    A DataHandler that's built for use with the G-matrix and V' terms
    from the 1999 Frederick and Woywood paper
    Usually used through the `KETermData` object.
    """

    base_dir = os.path.dirname(os.path.abspath(__file__))
    def __init__(self):
        super().__init__("KEData", data_dir=self.base_dir, data_pkg='PsiDatasets')#, alternate_keys=("Name", "Symbol", "CanonicalSymbol"))
    equivalent_perms = {
        "r":[[1, 0]],
        "a":[[2, 1, 0]],
        "t":[[3, 2, 1, 0]],
        "y":[]
    }
    def _get_remapping(self, i1, i2):
        remapping = {}
        n = 1
        _ = []
        for i in i1:
            if i not in remapping:
                remapping[i] = n
                _.append(n)
                n += 1
            else:
                _.append(remapping[i])
        i1 = tuple(_)

        _ = []
        for i in i2:
            if i not in remapping:
                remapping[i] = n
                _.append(n)
                n += 1
            else:
                _.append(remapping[i])
        i2 = tuple(_)

        return i1, i2
    def find_expressions(self, k, return_permutation=False):
        try:
            val = self[(k,)]
        except KeyError:
            pass
        else:
            if return_permutation:
                return val, None
            else:
                return val

        try:
            ((type1, type2), inds1, inds2) = k
        except:
            ((type1, type2), (inds1, inds2)) = k

        i1, i2 = self._get_remapping(inds1, inds2)
        k = ((type1, type2), i1, i2)
        try:
            val = self[(k,)]
        except KeyError:
            pass
        else:
            if return_permutation:
                return val, None
            else:
                return val


        for p1 in self.equivalent_perms[type1]:
            i1, i2 = self._get_remapping(tuple(inds1[j] for j in p1), inds2)
            k = ((type1, type2), i1, i2)
            try:
                val = self[(k,)]
            except KeyError:
                pass
            else:
                if return_permutation:
                    return val, (p1, None)
                else:
                    return val

            for p2 in self.equivalent_perms[type2]:
                i1, i2 = self._get_remapping(inds1, tuple(inds2[j] for j in p2))
                k = ((type1, type2), i1, i2)
                try:
                    val = self[(k,)]
                except KeyError:
                    pass
                else:
                    if return_permutation:
                        return val, (None, p2)
                    else:
                        return val

                i1, i2 = self._get_remapping(tuple(inds1[j] for j in p1), tuple(inds2[j] for j in p2))
                k = ((type1, type2), i1, i2)
                try:
                    val = self[(k,)]
                except KeyError:
                    pass
                else:
                    if return_permutation:
                        return val, (p1, p2)
                    else:
                        return val

        if return_permutation:
            return [0, 0], None
        else:
            return [0, 0]

    # def load(self):
    #     # now update by max equivalent indices
    #     super().load()
    #     new_vals = {}
    #
    #     self._data.update(new_vals)
KEData = KEDataHandler()
KEData.__doc__ = """An instance of KEDataHandler that can be used for looking up G-matrix and V' data"""
KEData.__name__ = "KEData"


"""
Provides analyzer class to handle common VPT analyses
"""

import enum, weakref, functools, numpy as np, itertools as ip, io
import os.path
import tempfile

from McUtils.Data import UnitsData
import McUtils.Numputils as nput
from McUtils.Scaffolding import Checkpointer, Logger, LogParser
from ..Spectra import DiscreteSpectrum
from .Wavefunctions import PerturbationTheoryWavefunctions
from .Runner import VPTRunner

__all__ = [
    "VPTResultsLoader",
    "VPTResultsSource",
    "VPTAnalyzer"
]

__reload_hook__ = ["..Spectra", ".Runner", ".Wavefunctions"]

class VPTResultsSource(enum.Enum):
    """
    Enum of sources to load PT results from
    """
    Wavefunctions = "wavefunctions"
    Checkpoint = "checkpoint"

def property_dispatcher(basefn):
    registry = weakref.WeakKeyDictionary()
    def register(key):
        if not isinstance(key, VPTResultsSource):
            key = VPTResultsSource(key)
        def bind(fn, key=key):
            fn.__name__ = basefn.__name__
            fn.__doc__ = basefn.__doc__
            registry[key] = fn
            return fn
        return bind

    @functools.wraps(basefn)
    def dispatch(self, *args, **kwargs):
        return registry[self.res_type](self, *args, **kwargs)
    dispatch.register = register

    return dispatch

class VPTResultsLoader:
    """
    Provides tools for loading results into canonical
    sources from a simulation, both from checkpoint files and from
    `PerturbationTheoryWavefunctions` objects and potentially more
    """

    def __init__(self, res, res_type=None):
        """
        :param res:
        :type res:
        :param res_type:
        :type res_type:
        """
        if isinstance(res, str):
            res = Checkpointer.from_file(res)
        self.data = res
        self.res_type = self.get_res_type(res) if res_type is None else res_type

    def get_res_type(self, res):
        """
        :param res:
        :type res:
        :return:
        :rtype:
        """
        if isinstance(res, PerturbationTheoryWavefunctions):
            return VPTResultsSource.Wavefunctions
        elif isinstance(res, Checkpointer):
            return VPTResultsSource.Checkpoint
        else:
            raise ValueError("do not know how to load PT results from {}".format(
                res
            ))

    @property_dispatcher
    def potential_terms(self):
        """
        Returns the expansion of the potential

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @potential_terms.register("checkpoint")
    def _(self):
        return self.data["potential_terms"]
    @potential_terms.register("wavefunctions")
    def _(self):
        return self.data.corrs.hamiltonian.V_terms

    @property_dispatcher
    def kinetic_terms(self):
        """
        Returns the expansion of the kinetic energy

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @kinetic_terms.register("checkpoint")
    def _(self):
        try:
            corr = self.data["coriolis_terms"]
        except KeyError:
            corr = None
        try:
            u = self.data["pseudopotential_terms"]
        except KeyError:
            u = None
        return {
            "G_matrix": self.data["gmatrix_terms"],
            "coriolis": corr,
            "pseudopotential": u
        }
    @kinetic_terms.register("wavefunctions")
    def _(self):
        return {
            "G_matrix": self.data.corrs.hamiltonian.G_terms,
            "coriolis": self.data.corrs.hamiltonian.coriolis_terms,
            "pseudopotential": self.data.corrs.hamiltonian.coriolis_terms
        }

    @property_dispatcher
    def dipole_terms(self):
        """
        Returns the expansion of the dipole moment

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @dipole_terms.register("checkpoint")
    def _(self):
        return self.data["dipole_terms"]
    @dipole_terms.register("wavefunctions")
    def _(self):
        return self.data.dipole_terms

    @property_dispatcher
    def basis(self):
        """
        Returns the basis for the calculation

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @basis.register("checkpoint")
    def _(self):
        return self.data["corrections"]["total_states"]
    @basis.register("wavefunctions")
    def _(self):
        data = self.data #type:PerturbationTheoryWavefunctions
        return data.corrs.total_basis

    @property_dispatcher
    def target_states(self):
        """
        Returns the target states for the calculation

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @target_states.register("checkpoint")
    def _(self):
        return self.data["corrections"]["states"]
    @target_states.register("wavefunctions")
    def _(self):
        data = self.data  # type:PerturbationTheoryWavefunctions
        return data.corrs.states.excitations

    @property_dispatcher
    def spectrum(self):
        """
        Returns the IR spectrum calculated from perturbation theory

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @spectrum.register("checkpoint")
    def _(self):
        freq, ints = self.data["spectrum"]
        return DiscreteSpectrum(freq * UnitsData.convert("Hartrees", "Wavenumbers"), ints)
    @spectrum.register("wavefunctions")
    def _(self):
        return DiscreteSpectrum(self.data.frequencies() * UnitsData.convert("Hartrees", "Wavenumbers"), self.data.intensities[1:])

    @property_dispatcher
    def zero_order_spectrum(self):
        """
        Returns the zero-order IR spectrum calculated from perturbation theory

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @zero_order_spectrum.register("checkpoint")
    def _(self):
        freq, ints = self.data["zero_order_spectrum"]
        return DiscreteSpectrum(freq * UnitsData.convert("Hartrees", "Wavenumbers"), ints)
    @zero_order_spectrum.register("wavefunctions")
    def _(self):
        return DiscreteSpectrum(self.data.deperturbed_frequencies(order=0) * UnitsData.convert("Hartrees", "Wavenumbers"),
                                self.data.deperturbed_intensities_to_order(order=0)[1:]
                                )

    @property_dispatcher
    def energy_corrections(self):
        """
        Returns the corrections to the energies

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @energy_corrections.register("checkpoint")
    def _(self):
        return self.data["corrections"]["energies"]
    @energy_corrections.register("wavefunctions")
    def _(self):
        data = self.data #type: PerturbationTheoryWavefunctions
        return data.corrs.energy_corrs

    def energies(self):
        """

        :return:
        :rtype:
        """
        return np.sum(self.energy_corrections(), axis=1)

    @property_dispatcher
    def wavefunction_corrections(self):
        """
        Returns the corrections to the wavefunctions

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @wavefunction_corrections.register("checkpoint")
    def _(self):
        return self.data["corrections"]["wavefunctions"]
    @wavefunction_corrections.register("wavefunctions")
    def _(self):
        return self.data.corrs.wfn_corrections

    @property_dispatcher
    def transition_moment_corrections(self):
        """
        Returns the corrections to the wavefunctions

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @transition_moment_corrections.register("checkpoint")
    def _(self):
        return self.data["transition_moments"]
    @transition_moment_corrections.register("wavefunctions")
    def _(self):
        return self.data.transition_moment_corrections
    def transition_moments(self):
        transition_moment_components = self.transition_moment_corrections()
        order = len(transition_moment_components[0])
        tmom = np.array([
            sum(
                sum(ip.chain(transition_moment_components[i][j]))
                if not isinstance(transition_moment_components[i][j], (float, int, np.floating, np.integer))
                else transition_moment_components[i][j]
                for i in range(order)  # correction order
            ) for j in range(3)  # xyz
        ])

        return tmom

    @property_dispatcher
    def deperturbed_transition_moment_corrections(self):
        """
        Returns the corrections to the wavefunctions

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @deperturbed_transition_moment_corrections.register("checkpoint")
    def _(self):
        return self.data["nondegenerate_transition_moments"]
    @deperturbed_transition_moment_corrections.register("wavefunctions")
    def _(self):
        return self.data.deperturbed_transition_moment_corrections
    def deperturbed_transition_moments(self):
        """
        :return:
        :rtype:
        """
        transition_moment_components = self.deperturbed_transition_moment_corrections()
        order = len(transition_moment_components[0])
        tmom = np.array([
            sum(
                sum(ip.chain(transition_moment_components[i][j]))
                if not isinstance(transition_moment_components[i][j], (float, int, np.floating, np.integer))
                else transition_moment_components[i][j]
                for i in range(order)  # correction order
            ) for j in range(3)  # xyz
        ])

        return tmom

    @property_dispatcher
    def degenerate_states(self):
        """
        Returns the deperturbed states used to make the degenerate transform

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @degenerate_states.register("checkpoint")
    def _(self):
        return self.data["degenerate_data"]["states"]
    @degenerate_states.register("wavefunctions")
    def _(self):
        data = self.data  # type:PerturbationTheoryWavefunctions
        states = data.corrs.degenerate_states
        if states is not None:
            states = [x for x in data.corrs.degenerate_states if len(x) > 1]
        return states

    @property_dispatcher
    def deperturbed_hamiltonians(self):
        """
        Returns the deperturbed Hamiltonians used to make the degenerate transform

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @deperturbed_hamiltonians.register("checkpoint")
    def _(self):
        return self.data["degenerate_data"]["hamiltonians"]
    @deperturbed_hamiltonians.register("wavefunctions")
    def _(self):
        data = self.data  # type:PerturbationTheoryWavefunctions
        return data.corrs.degenerate_hamiltonians

    @property_dispatcher
    def degenerate_energies(self):
        """
        Returns the deperturbed states used to make the degenerate transform

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @degenerate_energies.register("checkpoint")
    def _(self):
        return self.data["degenerate_data"]["energies"]
    @degenerate_energies.register("wavefunctions")
    def _(self):
        data = self.data  # type:PerturbationTheoryWavefunctions
        return data.corrs.degenerate_energies

    @property_dispatcher
    def degenerate_rotations(self):
        """
        Returns the deperturbed states used to make the degenerate transform

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @degenerate_rotations.register("checkpoint")
    def _(self):
        return self.data["degenerate_states"]["rotations"]
    @degenerate_rotations.register("wavefunctions")
    def _(self):
        data = self.data  # type:PerturbationTheoryWavefunctions
        return data.corrs.degenerate_transf

    @property_dispatcher
    def log_file(self):
        """
        Returns the log_file for the run

        :return:
        :rtype:
        """
        raise ValueError("no dispatch")
    @log_file.register("checkpoint")
    def _(self):
        tf = self.data.checkpoint_file
        tf, _ = os.path.splitext(tf)
        tf = tf+'.txt' #could be more sophisticated in the future maybe...
        if not os.path.exists(tf):
            tf = None
        return tf
    @log_file.register("wavefunctions")
    def _(self):
        return self.data.logger.log_file

def loaded_prop(fn):
    name = fn.__name__
    fn.__doc__ = getattr(VPTResultsLoader, name).__doc__
    return property(fn)
class VPTAnalyzer:
    """
    Provides analysis tools on VPT results
    """

    def __init__(self, res):
        """

        :param res:
        :type res:
        """
        if not isinstance(res, VPTResultsLoader):
            res = VPTResultsLoader(res)
        self.loader = res

    @classmethod
    def run_VPT(cls, *args, logger=None, **kwargs):
        """
        Runs a VPT calculation through `VPTRunner.run_simple` and
        stores the output wave functions to use

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if logger is None:
            logger = Logger(io.StringIO())

        wfns = VPTRunner.run_simple(
            *args,
            logger=logger,
            **kwargs
        )
        return cls(wfns)

    @loaded_prop
    def potential_terms(self):
        """

        :return:
        :rtype:
        """
        return self.loader.potential_terms()
    @loaded_prop
    def kinetic_terms(self):
        """

        :return:
        :rtype:
        """
        return self.loader.kinetic_terms()
    @loaded_prop
    def dipole_terms(self):
        """

        :return:
        :rtype:
        """
        return self.loader.dipole_terms()

    @loaded_prop
    def basis(self):
        """

        :return:
        :rtype:
        """
        return self.loader.basis()
    @loaded_prop
    def target_states(self):
        """

        :return:
        :rtype:
        """
        return self.loader.target_states()

    @loaded_prop
    def spectrum(self):
        """

        :return:
        :rtype:
        """
        return self.loader.spectrum()
    @loaded_prop
    def energy_corrections(self):
        """

        :return:
        :rtype:
        """
        return self.loader.energy_corrections()
    @loaded_prop
    def energies(self):
        """

        :return:
        :rtype:
        """
        engs = self.degenerate_energies
        if engs is None:
            engs = self.loader.energies()
        return engs
    @property
    def frequencies(self):
        """

        :return:
        :rtype:
        """
        return self.energies[1:] - self.energies[0]

    @loaded_prop
    def zero_order_spectrum(self):
        """

        :return:
        :rtype:
        """
        return self.loader.zero_order_spectrum()
    @property
    def deperturbed_spectrum(self):
        """

        :return:
        :rtype:
        """
        freqs = self.deperturbed_frequencies
        # if np.max(freqs) < 1:
        #     freqs = freqs * UnitsData.convert("Hartrees", "Wavenumbers")
        tmom = self.deperturbed_transition_moments[:, 0]
        osc = np.linalg.norm(tmom, axis=0) ** 2
        units = UnitsData.convert("OscillatorStrength", "KilometersPerMole")
        ints = units * freqs * osc[1:]
        return DiscreteSpectrum(freqs * UnitsData.convert("Hartrees", "Wavenumbers"), ints)
    @property
    def deperturbed_frequencies(self):
        """

        :return:
        :rtype:
        """
        return self.deperturbed_energies[1:] - self.deperturbed_energies[0]
    @loaded_prop
    def wavefunction_corrections(self):
        """

        :return:
        :rtype:
        """
        return self.loader.transition_moment_corrections()
    @loaded_prop
    def transition_moment_corrections(self):
        """

        :return:
        :rtype:
        """
        return self.loader.transition_moment_corrections()
    @loaded_prop
    def transition_moments(self):
        """

        :return:
        :rtype:
        """
        return self.loader.transition_moments()
    @loaded_prop
    def deperturbed_transition_moment_corrections(self):
        """

        :return:
        :rtype:
        """
        return self.loader.deperturbed_transition_moment_corrections()
    @loaded_prop
    def deperturbed_transition_moments(self):
        """

        :return:
        :rtype:
        """
        return self.loader.deperturbed_transition_moments()
    @loaded_prop
    def deperturbed_hamiltonians(self):
        """

        :return:
        :rtype:
        """
        return self.loader.deperturbed_hamiltonians()
    @property
    def deperturbed_energies(self):
        """

        :return:
        :rtype:
        """
        return self.loader.energies()
    @loaded_prop
    def degenerate_states(self):
        """

        :return:
        :rtype:
        """
        return self.loader.degenerate_states()
    @loaded_prop
    def degenerate_energies(self):
        """

        :return:
        :rtype:
        """
        return self.loader.degenerate_energies()

    def shift_and_transform_hamiltonian(self, hams, shifts):
        """

        :param hams:
        :type hams:
        :param shifts:
        :type shifts:
        :return:
        :rtype:
        """

        shifts = np.asanyarray(shifts)
        if shifts.ndim == 1:
            shifts = np.diag(shifts)

        ham = sum(hams) + shifts
        deg_engs, deg_transf = np.linalg.eigh(ham)

        # ov_thresh = .5
        # for i in range(len(deg_transf)):
        #     max_ov = np.max(deg_transf[:, i] ** 2)
        #     if max_ov < ov_thresh:  # there must be a single mode that has more than 50% of the initial state character?
        #         logger.log_print(
        #             "    state {i} is more than 50% mixed",
        #             i=i
        #         )
        #     #     raise PerturbationTheoryException("mode {} is has no contribution of greater than {}".format(
        #     #         i, ov_thresh
        #     #     ))

        # print(
        #     *str(
        #         np.round(ham * UnitsData.convert("Hartrees", "Wavenumbers")).astype(int)
        #     ).splitlines(),
        #     sep="\n"
        # )

        # we pick the terms with the max contribution from each input state
        # and zero out the contributions so that two states can't map
        # to the same input state
        sort_transf = np.abs(deg_transf.copy())
        sorting = [-1] * len(deg_transf)
        for i in range(len(deg_transf)):
            o = np.argmax(sort_transf[i, :])
            sorting[i] = o
            sort_transf[:, o] = 0.  # np.zeros(len(sort_transf))
        # print(deg_transf)

        deg_engs = deg_engs[sorting,]
        deg_transf = deg_transf[:, sorting]

        return deg_engs, deg_transf

    def get_shifted_transformed_transition_moments(self, deg_states, target_states, hams, shifts, tmoms, handling_mode='transpose'):
        """

        :param deg_states:
        :type deg_states:
        :param target_states:
        :type target_states:
        :param hams:
        :type hams:
        :param shifts:
        :type shifts:
        :param tmoms:
        :type tmoms:
        :param handling_mode:
        :type handling_mode:
        :return:
        :rtype:
        """

        deg_e, deg_t = self.shift_and_transform_hamiltonian(hams, shifts)
        inds, _ = nput.find(target_states, deg_states)
        subtms = [tm[0][inds] for tm in tmoms]

        tf = deg_t
        if handling_mode == 'transpose':
            tf = deg_t.T

        return deg_e, deg_t, [np.dot(tf, tm) for tm in subtms]
        # full_tm =

    def get_shifted_transformed_spectrum(self, zpe, deg_states, target_states, hams, shifts, tmoms, handling_mode='transpose'):
        """

        :param zpe:
        :type zpe:
        :param deg_states:
        :type deg_states:
        :param target_states:
        :type target_states:
        :param hams:
        :type hams:
        :param shifts:
        :type shifts:
        :param tmoms:
        :type tmoms:
        :param handling_mode:
        :type handling_mode:
        :return:
        :rtype:
        """


        eng, deg_transf, deg_tmom = self.get_shifted_transformed_transition_moments(
            deg_states, target_states, hams, shifts, tmoms,
            handling_mode=handling_mode
            )
        osc = np.linalg.norm(deg_tmom, axis=0) ** 2
        units = UnitsData.convert("OscillatorStrength", "KilometersPerMole")
        freqs = (eng - zpe)
        ints = units * freqs * osc#[1:]

        return DiscreteSpectrum(freqs * UnitsData.convert("Hartrees", "Wavenumbers"), ints), deg_transf

    def shifted_transformed_spectrum(self, deg_states, hams, shifts, return_transformation=False, handling_mode='transpose'):
        """

        :param deg_states:
        :type deg_states:
        :param hams:
        :type hams:
        :param shifts:
        :type shifts:
        :param return_transformation:
        :type return_transformation:
        :param handling_mode:
        :type handling_mode:
        :return:
        :rtype:
        """

        spec, tf = self.get_shifted_transformed_spectrum(
            self.energies[0], deg_states, self.target_states, hams, shifts, self.deperturbed_transition_moments,
            handling_mode=handling_mode
            )
        if not return_transformation:
            return spec
        else:
            return spec, tf

    def transition_data(self, states, keys=['frequency', 'transition_moment', 'intensity'], data='deperturbed'):
        """

        :param states:
        :type states:
        :param keys:
        :type keys:
        :param data:
        :type data:
        :return:
        :rtype:
        """

        states = np.asanyarray(states)

        single = states.ndim == 1
        if single:
            states = np.expand_dims(states, 0)
        inds, _ = nput.find(self.target_states, states)

        freq = self.deperturbed_frequencies if data == 'deperturbed' else self.frequencies

        res = {'index':inds}
        if 'frequency' in keys:
            res['frequency'] = freq[inds-1] * UnitsData.convert("Hartrees", "Wavenumbers")
        if 'max_contribs' in keys:
            raise NotImplementedError("whoops")
            res['max_contribs'] = ...
        if 'transition_moment' in keys or 'intensity' in keys:
            if data == 'deperturbed':
                tm_base = self.deperturbed_transition_moments
            else:
                tm_base = self.transition_moments
            res['transition_moment'] = np.array([tm[0][inds] for tm in tm_base]).T
            res['intensity'] = freq[inds-1] * np.linalg.norm(res['transition_moment'], axis=1)**2 * UnitsData.convert("OscillatorStrength", "KilometersPerMole")

        new_res = []
        for i in range(len(states)):
            new_res.append({k:res[k][i] for k in res.keys()})

        if single:
            new_res = new_res[0]

        return new_res

    @staticmethod
    def _aggregate_tmoms(tmom_corrs, inds, terms):

        corrs = []
        order = len(tmom_corrs[0])
        for a in range(3):
            cur = 0
            for o in range(order):
                n = 0
                for i, j, k in ip.product(range(o + 1), range(o + 1), range(o + 1)):
                    if i + j + k == o:
                        if (i, k, j) in terms:
                            cur += tmom_corrs[o][a][n][0][inds]

                        n += 1
            corrs.append(cur)

        res = np.array(corrs).T
        return res

    def transition_moment_term_sums(self, states, terms=None, rotation=None, data='deperturbed'):
        """

        :param states:
        :type states:
        :param terms:
        :type terms:
        :param rotation:
        :type rotation:
        :param data:
        :type data:
        :return:
        :rtype:
        """
        if terms is None:
            terms = {
                'harmonic': [(0, 0, 0)],
                'electrical': [(0, 1, 0), (0, 2, 0)],
                'mechanical': [(1, 0, 0), (0, 0, 1), (1, 0, 1), (2, 0, 0), (0, 0, 2)],
                'mixed': [(1, 1, 0), (0, 1, 1)],
            }

        states = np.asanyarray(states)

        single = states.ndim == 1
        if single:
            states = np.expand_dims(states, 0)
        inds, _ = nput.find(self.target_states, states)

        tmom_corrs = self.deperturbed_transition_moment_corrections if data == 'deperturbed' else self.transition_moment_corrections

        res = {}
        for k,t in terms.items():
            res[k] = self._aggregate_tmoms(tmom_corrs, inds, t)

        new_res = []
        for i in range(len(states)):
            if rotation is None:
                new_res.append({k:res[k][i] for k in res.keys()})
            else:
                new_res.append({k:rotation@res[k][i] for k in res.keys()})

        if single:
            new_res = new_res[0]

        return new_res

    def transition_moment_term_sums_first_order(self, states, rotation=None, data='deperturbed'):
        """

        :param states:
        :type states:
        :param rotation:
        :type rotation:
        :param data:
        :type data:
        :return:
        :rtype:
        """
        return self.transition_moment_term_sums(
            states,
            terms = {
                    'harmonic': [(0, 0, 0)],
                    'electrical': [(0, 1, 0)],
                    'mechanical': [(1, 0, 0), (0, 0, 1)]
                },
            rotation=rotation,
            data=data
        )

    def intensity_breakdown(self, states, terms=None, data='deperturbed'):
        """

        :param states:
        :type states:
        :param terms:
        :type terms:
        :param data:
        :type data:
        :return:
        :rtype:
        """

        woop = self.transition_moment_term_sums(states, terms=terms, data=data)
        single = isinstance(woop, dict)
        if single:
            woop = [woop]

        breakdowns = [{k:np.linalg.norm(v)**2 for k,v in w.items()} for w in woop]

        if single:
            breakdowns = breakdowns[0]

        return breakdowns

    def degenerate_coupling_element(self, state1, state2):
        """

        :param state1:
        :type state1:
        :param state2:
        :type state2:
        :return:
        :rtype:
        """

        for g,h in zip(self.degenerate_states, self.deperturbed_hamiltonians):
            try:
                pos = nput.find(g, np.array([state1]))[0]
            except IndexError:
                continue
            else:
                pos2 = nput.find(g, np.array([state2]))[0]
                return sum(h)[pos[0], pos2[0]] * UnitsData.convert("Hartrees", "Wavenumbers")
        else:
            raise IndexError("not in any degenerate group")


    def format_deperturbed_hamiltonian(self, which):
        """

        :param which:
        :type which:
        :return:
        :rtype:
        """
        with np.printoptions(linewidth=1e9):
            return str(sum(self.deperturbed_hamiltonians[which]) * UnitsData.convert("Hartrees", "Wavenumbers"))

    @property
    def log_parser(self):
        log_file = self.loader.log_file()
        if isinstance(log_file, str):
            lf = log_file
        elif isinstance(log_file, io.StringIO):
            with tempfile.NamedTemporaryFile(delete=False) as tf:
                lf = tf.name
            with open(lf, 'w+') as tf:
                log_file.seek(0)
                tf.write(log_file.read())
        else:
            lf = None
        if lf is not None:
            return LogParser(lf)
        else:
            raise ValueError("log file {} can't be parsed".format(log_file))

    def print_output_tables(self,
                            print_energy_corrections=False,
                            print_energies=False,
                            print_transition_moments=False,
                            print_intensities=True,
                            **kwargs):
        no_print = False
        try:
            logger = self.loader.data.logger
        except AttributeError:
            no_print = True
            with self.log_parser:
                for block in self.log_parser.get_blocks():
                    if not isinstance(block, str) and block.tag == 'IR Data':
                        print(block)
        else:
            self.loader.data.logger = None
            VPTRunner.print_output_tables(
                self.loader.data,
                print_energy_corrections=print_energy_corrections,
                print_energies=print_energies,
                print_transition_moments=print_transition_moments,
                print_intensities=print_intensities,
                **kwargs
            )
        finally:
            if not no_print:
                self.loader.data.logger = logger


"""
Provides a direct-product extension to a system of 1D DVRs
"""

import numpy as np, scipy.sparse as sp
from functools import reduce
from .BaseDVR import BaseDVR
from .ColbertMiller import *

__all__ = [
    "DirectProductDVR",
    "CartesianNDDVR",
    "RingNDDVR",
    "SphericalDVR"
]

class DirectProductDVR(BaseDVR):
    def __init__(self,
                 dvrs_1D,
                 zero_threshold=1e-14,
                 **base_opts
                 ):
        """
        :param dvrs_1D: a series of 1D DVRs that can provide the inputs we'll product together
        :type dvrs_1D: Iterable[AbstractDVR]
        :param base_opts:
        :type base_opts:
        """
        super().__init__(**base_opts)
        self.zero_threshold=zero_threshold
        self.dvrs = dvrs_1D
    def __repr__(self):
        return "{}({}, pot={})".format(
            type(self).__name__,
            self.dvrs,
            self.potential_function
        )

    def get_grid(self, domain=None, divs=None, **kwargs):

        if domain is None:
            subgrids = [dvr.grid() for dvr in self.dvrs]
        else:
            subgrids = [dvr.grid(domain=dom, divs=div) for dvr,dom,div in zip(self.dvrs, domain, divs)]
        tfs = [g[1] if isinstance(g, tuple) and len(g) == 2 else None for g in subgrids] # FBR DVRs return a tf as well
        subgrids = [g[0] if isinstance(g, tuple) and len(g) == 2 else g for g in subgrids] # FBR DVRs return a tf as well
        mesh = np.array(np.meshgrid(*subgrids, indexing='ij'))
        MEHSH = np.moveaxis(mesh, 0, len(subgrids))

        if any(t is not None for t in tfs):
            return MEHSH, tfs
        else:
            return MEHSH

    def grid(self, domain=None, divs=None, **kwargs):
        if domain is None:
            domain = self.domain
        if divs is None:
            divs = self.divs

        # if divs is None:
        #     raise ValueError("need a value for `domain`")
        # if divs is None:
        #     raise ValueError("need a value for `divs`")

        return self.get_grid(domain=domain, divs=divs, **kwargs)

    def get_kinetic_energy(self, grid=None, mass=None, hb=1, g=None, g_deriv=None, logger=None, include_kinetic_coupling=True, **kwargs):

        if isinstance(grid, tuple) and len(grid) == 2:
            grid, tfs = grid
        else:
            tfs = None

        ndims = grid.shape[-1]
        try:
            iter(mass); ms = mass
        except TypeError:
            ms = [mass] * ndims

        try:
            iter(hb); hbs = hb
        except TypeError:
            hbs = [hb] * ndims

        ndim = grid.shape[-1]
        grids = [  # build subgrids
            grid[(0,) * i + (...,) + (0,) * (ndim - i - 1) + (i,)]
            for i in range(ndim)
        ]
        if g is not None:
            if g_deriv is None:
                raise ValueError(
                    "if functions for `g` are supplied, also need functions, `g_deriv` for the second derivative of `g`")

            include_coupling = include_kinetic_coupling and any(
                i != j
                and not (
                        isinstance(g[i][j], (int, float, np.integer, np.floating))
                        and abs(g[i][j]) < self.zero_threshold
                )
                for i in range(ndim) for j in range(ndim)
            )

            ms = [1] * len(ms)

        else:
            include_coupling = False

        if tfs is None:
            tfs = [None]*len(ms)
        kes = [
            dvr.kinetic_energy((subg, tf) if tf is not None else subg, mass=m, hb=hb)
            for dvr, subg, tf, m, hb in zip(self.dvrs, grids, tfs, ms, hbs)
        ]
        kes = [sp.csr_matrix(mat) for mat in kes]
        if g is None:  # we passed constant masses
            logger.log_print('no kinetic coupling')
            def _kron_sum(a, b):
                '''Computes a Kronecker sum to build our Kronecker-Delta tensor product expression'''
                n_1 = a.shape[0]
                n_2 = b.shape[0]
                ident_1 = sp.identity(n_1)
                ident_2 = sp.identity(n_2)

                return sp.kron(a, ident_2) + sp.kron(ident_1, b)

            ke = reduce(_kron_sum, kes)
        else:
            flat_grid = np.reshape(grid, (-1, ndim))
            tot_shape = [len(gr) for gr in grids]  # we'll need this to multiply terms into the right indices
            ke = sp.csr_matrix((len(flat_grid), len(flat_grid)), dtype=kes[0].dtype)
            for i in range(ndim):  # build out all of the coupling term products
                # evaluate g over the terms and average
                if not (
                        isinstance(g[i][i], (int, float, np.integer, np.floating))
                        and abs(g[i][i]) < self.zero_threshold
                ):
                    if isinstance(g[i][i], (int, float, np.integer, np.floating)):
                        g_vals = np.full(grid.shape[:-1], g[i][i])
                    else:
                        g_vals = np.reshape(g[i][i](flat_grid), grid.shape[:-1])

                    # construct the basic kinetic energy kronecker product
                    sub_kes = [  # set up all the subtensors we'll need for this
                        sp.eye(tot_shape[k]) if k != i else kes[k] for k in range(ndim)
                    ]
                    ke_mat = reduce(sp.kron, sub_kes)

                    # now we need to figure out where to multiply in the g_vals
                    flat_rows, flat_cols, ke_vals = sp.find(ke_mat)
                    # we convert each row and column into its corresponding direct
                    # product index since each is basically a flat index for a multdimensional
                    # array
                    row_inds = np.unravel_index(flat_rows, tot_shape)
                    col_inds = np.unravel_index(flat_cols, tot_shape)

                    # and we pull the G matrix values for the corresponding i and j indices
                    row_vals = g_vals[row_inds]
                    col_vals = g_vals[col_inds]

                    # finally we take the average of the two and put them into a sparse matrix
                    # that can be multiplied by the base kinetic matrix values
                    avg_g_vals = 1 / 2 * (row_vals + col_vals)
                    ke_mat = sp.csr_matrix(
                        (
                            avg_g_vals * ke_vals,
                            (flat_rows, flat_cols)
                        ),
                        shape=ke.shape,
                        dtype=ke_vals.dtype
                    )

                    ke += ke_mat

                    if isinstance(g_deriv[i], (int, float, np.integer, np.floating)):
                        gd_vals = np.full(flat_grid.shape[:-1], g_deriv[i])
                    else:
                        gd_vals = g_deriv[i](flat_grid)

                    gd_vals = 1 / 4 * gd_vals
                    ke_contrib = sp.diags([gd_vals], [0])
                    ke += ke_contrib

            if include_coupling:
                logger.log_print('evaluating kinetic coupling')
                momenta = [dvr.real_momentum((subg, tf) if tf is not None else subg, hb=hb) for dvr,subg,tf in zip(self.dvrs, grids, tfs)]
                kinetic_coupling = sp.csr_matrix(ke.shape, dtype=ke.dtype)  # initialize empty tensor
                for i in range(len(momenta)):  # build out all of the coupling term products
                    for j in range(i + 1, len(momenta)):
                        if i != j and not (isinstance(g[i][j], (int, float, np.integer, np.floating)) and abs(g[i][j]) < self.zero_threshold):

                            if isinstance(g[i][j], (int, float, np.integer, np.floating)):
                                g_vals = np.full(grid.shape[:-1], g[i][j])
                            else:
                                g_vals = np.reshape(g[i][j](flat_grid), grid.shape[:-1])

                            # construct the basic momenta Kroenecker product
                            sub_momenta = [  # set up all the subtensors we'll need for this
                                sp.eye(tot_shape[k]) if k != i and k != j else sp.csr_matrix(momenta[k])
                                for k in range(len(momenta))
                            ]
                            momentum_mat = reduce(sp.kron, sub_momenta)

                            # now we need to figure out where to multiply in the ij_vals
                            flat_rows, flat_cols, mom_prod_vals = sp.find(momentum_mat)

                            # we convert each row and column into its corresponding direct
                            # product index since each is basically a flat index for a multdimensional
                            # array
                            row_inds = np.unravel_index(flat_rows, tot_shape)
                            col_inds = np.unravel_index(flat_cols, tot_shape)

                            # we do the swap of row/column indices required by the mixed term
                            swap_rows = row_inds[:i] + (col_inds[i],) + row_inds[i+1:]
                            swap_cols = col_inds[:i] + (row_inds[i],) + col_inds[i+1:]

                            # and we pull the G matrix values for the corresponding indices
                            row_vals = g_vals[swap_rows]
                            col_vals = g_vals[swap_cols]

                            # finally we take the sum of the two and put them into a sparse matrix
                            # that can be multiplied by the base momentum matrix values
                            sum_g_vals = (row_vals + col_vals)
                            coupling_vals = 1/2 * sum_g_vals * mom_prod_vals
                            mask = np.abs(coupling_vals) > self.zero_threshold
                            coupling_vals = coupling_vals[mask]
                            flat_rows = flat_rows[mask]
                            flat_cols = flat_cols[mask]
                            coupling_term = sp.csr_matrix(
                                (
                                    coupling_vals,
                                    (flat_rows, flat_cols)
                                ),
                                shape=momentum_mat.shape,
                                dtype=momentum_mat.dtype
                            )

                            kinetic_coupling -= coupling_term  # negative sign from the two factors of i
                ke += kinetic_coupling
                if ke.getnnz() >= 1 / 2 * np.prod(ke.shape):
                    ke = ke.toarray()

        return ke

    def kinetic_energy(self, grid=None, mass=None, hb=1, g=None, g_deriv=None, logger=None, **kwargs):
        """
        Computes the N-dimensional kinetic energy
        :param grid:
        :type grid:
        :param mass:
        :type mass:
        :param hb:
        :type hb:
        :param g:
        :type g:
        :param g_deriv:
        :type g_deriv:
        :return:
        :rtype:
        """

        if grid is None:
            grid = self.grid()

        logger = self._logger(logger)
        return self.get_kinetic_energy(grid=grid, mass=mass, hb=hb, g=g, g_deriv=g_deriv, logger=logger, **kwargs)

class CartesianNDDVR(DirectProductDVR):
    """
    Provides an ND-DVR over different domains
    """
    def __init__(self,
                 domains,
                 **base_opts
                 ):

        sub_dvrs = [CartesianDVR(domain=dom[:2], divs=dom[2]) for dom in domains]
        super().__init__(
            sub_dvrs,
            **base_opts
        )

class RingNDDVR(DirectProductDVR):
    """
    Provides an ND-DVR for products of periodic (0, 2Pi) ranges
    """

    def __init__(self,
                 divs,
                 **base_opts
                 ):
        sub_dvrs = [RingDVR(domain=(0, 2*np.pi), divs=div) for div in divs]
        super().__init__(
            sub_dvrs,
            **base_opts
        )

class SphericalDVR(DirectProductDVR):
    def __init__(self,
                 r_max,
                 divs,
                 **base_opts
                 ):
        sub_dvrs = [
            RadialDVR(domain=(0, r_max), divs=divs[0]),
            RingDVR(domain=(0, 2*np.pi), divs=divs[1]),
            PolarDVR(domain=(0, 2*np.pi), divs=divs[1]),


        ]
        super().__init__(
            sub_dvrs,
            **base_opts
        )





import numpy as np
from ..BasisReps import PermutationStateIndexer
from .Wavefunctions import Wavefunctions, Wavefunction

__all__ = [
    "DirectProductWavefunction",
    "DirectProductWavefunctions",
]
class DirectProductWavefunction(Wavefunction):
    def __init__(self, wfns:'Iterable[Wavefunction]', parent=None, index=None, **opts):
        energies = [w.energy for w in wfns]
        self.energies = energies
        super().__init__(
            sum(energies),
            wfns,
            parent=parent,
            index=index,
            **opts
        )

class DirectProductWavefunctions(Wavefunctions):
    wavefunction_class = DirectProductWavefunction
    def __init__(self,
                 wfns,
                 indices=None, wavefunction_class=None, **opts):
        self.base_wfns = wfns
        self.indexer = PermutationStateIndexer(len(wfns))
        super().__init__(
            None,
            wfns,
            indices=indices,
            wavefunction_class=wavefunction_class,
            **opts
        )

    def get_slice(self, n):
        if isinstance(n, int):
            ...
        else:
            np.arange()
        self.indexer.from_indices(len)

    def _get_slice_inds(self, n):
        if isinstance(n, int):
            idx = np.arange(n)
        elif n is None:
            idx = np.arange(len(self))
        elif isinstance(n, slice):
            stop = n.stop
            if stop is None:
                stop = len(self)
            idx = np.arange(stop)[n]
        else:
            idx = n
        return self.indexer.from_indices(idx)

    def __len__(self):
        return np.prod([len(w) for w in self.base_wfns]).astype(int)

    def get_energies(self, idx):
        base_idx = self._get_slice_inds(idx).T
        return sum(w.energies[i] for w,i in zip(self.base_wfns, base_idx))
    @property
    def energies(self):
        return self.get_energies(None)

    def get_wavefunctions(self, which):
        base_idx = self._get_slice_inds(which).T
        wfn_bits = [w.wavefunctions[i] for w, i in zip(self.base_wfns, base_idx)]
        if isinstance(wfn_bits[0], Wavefunction):
            return self.wavefunction_class(wfn_bits, parent=self, index=which)
        else:
            return type(self)(wfn_bits, indices=base_idx.T)